import datetime
import math
import uuid
from copy import copy
from decimal import Decimal
from typing import List, Union, Any
from unittest.mock import ANY

import pandas as pd
from sqlalchemy.engine import Engine

from bach import DataFrame, Series
from sql_models.util import is_postgres, is_bigquery, is_athena


def _get_view_sql_data(df: DataFrame):
    sql = df.view_sql()
    # escape sql, as conn.execute will think that '%' indicates a parameter
    sql = sql.replace('%', '%%')
    with df.engine.connect() as conn:
        db_rows = conn.execute(sql)

    column_names = list(db_rows.keys())
    db_values = [list(row) for row in db_rows]
    print(db_values)
    return column_names, db_values


def _get_to_pandas_data(df: DataFrame):
    pdf = df.to_pandas()
    # Convert pdf to the same format as _get_view_sql_data gives
    column_names = (list(pdf.index.names) if df.index else []) + list(pdf.columns)
    pdf = pdf.reset_index()
    db_values = []
    for value_row in pdf.to_numpy().tolist():
        db_values.append(value_row if df.index else value_row[1:])  # don't include default index value
    print(db_values)
    return column_names, db_values


def _convert_uuid_expected_data(engine: Engine, data: List[List[Any]]) -> List[List[Any]]:
    """
    Convert any UUID objects in data to string, if we represent uuids with strings in the engine's dialect.
    """
    if is_postgres(engine):
        return data
    if is_athena(engine) or is_bigquery(engine):
        result = [
            [str(cell) if isinstance(cell, uuid.UUID) else cell for cell in row]
            for row in data
        ]
        return result
    raise Exception(f'engine not supported {engine}')


def _is_na(value: Any) -> bool:
    """
    Check if value is NaN, Nat, or None.
    Similar to pandas.isna(), but won't raise a warning when value is an empty array
    """
    is_nan = isinstance(value, float) and math.isnan(value)
    is_nat = value is pd.NaT
    is_none = value is None
    return is_nan or is_nat or is_none


def assert_equals_data(
    bt: Union[DataFrame, Series],
    expected_columns: List[str],
    expected_data: List[list],
    order_by: Union[str, List[str]] = None,
    use_to_pandas: bool = False,
    round_decimals: bool = False,
    decimal=4,
    convert_uuid: bool = False,
) -> List[List[Any]]:
    """
    Execute the sql of DataFrame/Series's view_sql(), with the given order_by, and make sure the
    result matches the expected columns and data.

    Note: By default this does not call `to_pandas()`, which we nowadays consider our 'normal' path,
    but directly executes the result from `view_sql()`. To test `to_pandas()` set use_to_pandas=True.
    :return: the values queried from the database
    """
    if len(expected_data) == 0:
        raise ValueError("Cannot check data if 0 rows are expected.")

    if isinstance(bt, Series):
        # Otherwise sorting does not work as expected
        bt = bt.to_frame()

    if convert_uuid:
        expected_data = _convert_uuid_expected_data(bt.engine, expected_data)

    if order_by:
        bt = bt.sort_values(order_by)
    elif not bt.order_by:
        bt = bt.sort_index()

    if not use_to_pandas:
        column_names, db_values = _get_view_sql_data(bt)
    else:
        column_names, db_values = _get_to_pandas_data(bt)

    assert len(db_values) == len(expected_data), f'{len(db_values)} != {len(expected_data)}'
    assert column_names == expected_columns, f'{column_names} != {expected_columns}'

    _date_freq = 'ms' if is_athena(bt.engine) else 'us'
    for i, df_row in enumerate(db_values):
        expected_row = expected_data[i]
        assert len(df_row) == len(expected_row)
        for j, val in enumerate(df_row):
            actual = copy(val)
            expected = copy(expected_row[j])

            if isinstance(expected, (float, Decimal)) and round_decimals:
                actual = round(Decimal(actual), decimal)
                expected = round(Decimal(expected), decimal)

            if isinstance(actual, (datetime.date, pd.Timestamp)):
                actual = pd.Timestamp(actual)
                actual = actual.floor(freq=_date_freq)

                if isinstance(expected, (pd.Timestamp, datetime.datetime, datetime.date)):
                    pdt_expected = pd.Timestamp(expected)
                    expected = pdt_expected.floor(freq=_date_freq)

            assert_msg = f'row {i} is not equal: {expected_row} != {df_row}; value: {expected} != {actual}'
            if not _is_na(actual):
                assert actual == expected, assert_msg
            elif type(expected) is type(ANY):
                # Actual is None, NaN, or NaT, and expected is ANY. This is OK
                # We handle this as a separate case because isna() and isnan() functions don't understand
                # ANY
                pass
            else:
                # Generically check: if actual is None/NaN/NaT, then expected should be None/NaN/NaT too
                assert _is_na(expected), assert_msg
                # If we specifically expect NaN, then we expect the result to be NaN
                # If we specifically expect NaT, then we expect the result to be NaT
                # If we expect None, then the result can be any of None, Nan, or NaT
                if isinstance(expected, float) and math.isnan(expected):
                    assert isinstance(actual, float), assert_msg
                    assert math.isnan(actual), assert_msg
                elif expected is pd.NaT:
                    assert actual is pd.NaT, assert_msg
    return db_values



"""
Copyright 2021 Objectiv B.V.

Functions for looking up the right classes to handle types and values, and for registering new custom
types.

To prevent cyclic imports, the functions in this file should not be used by dataframe.py before the file
is fully initialized (that is, only use within functions).
"""
from copy import copy
from typing import Type, Tuple, Any, TypeVar, List, TYPE_CHECKING, Dict, Union, Sequence, Mapping, Set
import datetime
from uuid import UUID

import numpy
import pandas

from sql_models.constants import DBDialect

if TYPE_CHECKING:
    from bach.series import Series


AllSupportedLiteralTypes = Union[
    int, numpy.int64,
    float, numpy.float64,
    bool,
    str,
    None,
    datetime.date, datetime.time, datetime.datetime, numpy.datetime64, datetime.timedelta, numpy.timedelta64,
    UUID,
    pandas.Interval,
    dict,
    list
]
"""
AllSupportedLiteralTypes are all the types for which a Series is registered to interpret the literal.
Of course when custom types are added, this definition will be incomplete, but as this is just for mypy
usage, that is fine.
"""


Dtype = str
StructuredDtype = Union[List[Any], Mapping[str, Any], Dtype]
# Real definition of StructuredDtype, but not supported by mypy:
# StructuredDtype = Union[List[StructuredDtype], Dict[str, StructuredDtype], str]
DtypeOrAlias = Union[Type, StructuredDtype]


def get_series_type_from_dtype(dtype: DtypeOrAlias) -> Type['Series']:
    """ Given a dtype, return the correct Series subclass. """
    return _registry.get_series_type_from_dtype(dtype)


def get_dtype_from_db_dtype(db_dialect: DBDialect, db_dtype: str) -> StructuredDtype:
    """ Given a database datatype, return the dtype of the Series subclass for that datatype. """
    return _registry.get_dtype_from_db_dtype(db_dialect=db_dialect, db_dtype=db_dtype)


def value_to_dtype(value: Any) -> Dtype:
    """ Give the dtype, as a string of the given value. """
    return _registry.value_to_dtype(value)


def get_series_type_from_db_dtype(db_dialect: DBDialect, db_dtype: str) -> Type['Series']:
    """ Given a database datatype, return the correct Series subclass. """
    dtype = get_dtype_from_db_dtype(db_dialect=db_dialect, db_dtype=db_dtype)
    return get_series_type_from_dtype(dtype)


def value_to_series_type(value: Any) -> Type['Series']:
    """ Return the Series subclass that can represent value as literal. """
    return get_series_type_from_dtype(dtype=value_to_dtype(value))


def get_all_db_dtype_to_series() -> Mapping[DBDialect, Mapping[Dtype, Type['Series']]]:
    """ Get all registered mappings from db_dtype to Series type"""
    return _registry.get_all_db_dtype_to_series()


def validate_is_dtype(dtype: StructuredDtype):
    """
    Validate that the given dtype is a valid dtype. Raises a ValueError otherwise.
    Note that here we don't consider dtype-aliasses to be valid.
    """
    return _registry.validate_is_dtype(dtype)


def validate_instance_dtype(static_dtype: Dtype, instance_dtype: StructuredDtype):
    """
    Validate that:
    1) instance_dtype is a valid instance dtype for the given static_dtype.
    2) instance_dtype is well-formed.

    Example: `['int64', 'float64']` is not a well formed instance-dtype. a list should have only one item.
    Example: `['int64']` is not well formed instance-dtype, and a correct instance_dtype
                if static_dtype='list', but not when static_dtype='bool'

    :raises ValueError: if any check fails
    """
    # hardcoded for now, make more elegant in the future when we have a lot of DBs to support
    structural_type_mapping: Dict[Dtype, Type] = {
        'list': list,
        'dict': dict
    }
    expected_type = structural_type_mapping.get(static_dtype)
    if expected_type is not None:
        if not isinstance(instance_dtype, expected_type):
            raise ValueError(f'Expected instance dtype of type "{expected_type}". '
                             f'instance_dtype type: {type(instance_dtype)}, '
                             f'instance_dtype: {instance_dtype}')
    else:
        if static_dtype != instance_dtype:
            raise ValueError(f'Expected instance dtype "{static_dtype}". '
                             f'instance_dtype: "{instance_dtype}"')

    return validate_is_dtype(instance_dtype)


def validate_dtype_value(
    static_dtype: Dtype,
    instance_dtype: StructuredDtype,
    value: Union['Series', AllSupportedLiteralTypes]
):
    """
    Check that static_dtype is a valid instance dtype for the given static_dtype, and that value is of the
    given instance-dtype.

    :param static_dtype: static dtype against which to validate the instance_dtype
    :param instance_dtype: instance dtype against which to validate value.
    :param value: Value to validate. Can either be a python value or a Series.
    :raises ValueError: if any checks fail
    """
    validate_instance_dtype(static_dtype=static_dtype, instance_dtype=instance_dtype)
    _validate_dtype_of_value(dtype=instance_dtype, value=value)


T = TypeVar('T', bound='Series')


def register_dtype(value_types: List[Type] = None, override_registered_types: bool = False):
    """
    Decorator to register a Series subclass as dtype series
    :value_types: List of Types for which values should be instantiated as the registered class
    :override_registerd_types: If False an Exception is raised if the class' dtype or db_dtype conflict
        with existing registerd types or if one of the value_types are already coupled. If True this new
        registration will override the existing one.
    """
    if value_types is None:
        value_types = []

    def wrapper(cls: Type[T]) -> Type[T]:
        # Mypy needs some help here
        assert value_types is not None
        _registry.register_dtype_series(cls, value_types, override_registered_types)
        return cls
    return wrapper


class TypeRegistry:
    def __init__(self):
        # Do the real initialisation in _real_init, which we'll defer until usage so we won't get
        # problems with cyclic imports.

        # dtype_series: Mapping of dtype and dtype-alias to a subclass of Series
        self.dtype_to_series: Dict[DtypeOrAlias, Type['Series']] = {}

        # db_dtype_to_series: Mapping per database dialect, of database types to a subclass of Series
        self.db_dtype_to_series: Dict[DBDialect, Dict[Dtype, Type['Series']]] = {}

        # value_type_to_series: Mapping of python types to matching Series types
        # note that some types could be in this dictionary multiple times. For a subtype its super types
        # might also be in the list. We resolve conflicts in arg_to_type by returning the latest matching
        # entry.
        # This is also the reason this is a list of typles instead of a dictionary: the order is important
        # and that is clearer with a list.
        self.value_type_to_series: List[Tuple[Type, Type['Series']]] = []

    def _real_init(self):
        """
        Load the default dtype, db_dtype, and value-type mappings for the standard set of Series types.

        The dtype_to_series mapping will be based on the dtype and dtype_aliases that the standard
            Series declare.
        The db_dtype_to_series mapping will be similarly based on the supported_db_dtype that the standard
            Series declare.
        The value_type_to_series is hardcoded here for the standard Series.
        """
        if self.dtype_to_series or self.db_dtype_to_series or self.value_type_to_series:
            # Only initialise once
            return

        # Import locally to prevent cyclic imports
        from bach.series import \
            SeriesBoolean, SeriesInt64, SeriesFloat64, SeriesString,\
            SeriesTimestamp, SeriesDate, SeriesTime, SeriesTimedelta,\
            SeriesUuid, SeriesJson, SeriesJsonPostgres, SeriesNumericInterval, SeriesList, SeriesDict

        standard_types: List[Type[Series]] = [
            SeriesBoolean, SeriesInt64, SeriesFloat64, SeriesString,
            SeriesTimestamp, SeriesDate, SeriesTime, SeriesTimedelta,
            SeriesUuid, SeriesJson, SeriesJsonPostgres, SeriesNumericInterval, SeriesList, SeriesDict
        ]

        for klass in standard_types:
            self._register_dtype_klass(klass)
            self._register_db_dtype_klass(klass)

        # For the value_type_to_dtype list order can be important. The value_to_dtype() function starts at
        # end of this list and the first matching entry determines the return value.
        # A type might match multiple entries in the list, because it can be an instance of multiple (super)
        # classes. E.g. a `bool` is also an `int`
        # Therefore this list is hardcoded here, and not automatically derived from the base_types classes
        # When adding an item here, make sure to update AllSupportedLiteralTypes above
        self._register_value_klass(int, SeriesInt64)
        self._register_value_klass(numpy.int64, SeriesInt64)
        self._register_value_klass(float, SeriesFloat64)
        self._register_value_klass(numpy.float64, SeriesFloat64)
        self._register_value_klass(bool, SeriesBoolean)
        self._register_value_klass(str, SeriesString)
        self._register_value_klass(type(None), SeriesString)  # NoneType defaults to String.
        self._register_value_klass(pandas.Interval, SeriesNumericInterval)

        self._register_value_klass(datetime.date, SeriesDate)
        self._register_value_klass(datetime.time, SeriesTime)
        self._register_value_klass(datetime.datetime, SeriesTimestamp)
        self._register_value_klass(numpy.datetime64, SeriesTimestamp)
        self._register_value_klass(datetime.timedelta, SeriesTimedelta)
        self._register_value_klass(numpy.timedelta64, SeriesTimedelta)
        self._register_value_klass(UUID, SeriesUuid)
        self._register_value_klass(dict, SeriesJson)
        self._register_value_klass(list, SeriesJson)

    def _register_dtype_klass(self, klass: Type['Series'], override=False):
        klass_dtype: DtypeOrAlias = klass.dtype
        dtype_and_aliases: Sequence[DtypeOrAlias] = [klass_dtype] + list(klass.dtype_aliases)
        for dtype_alias in dtype_and_aliases:
            if dtype_alias in self.dtype_to_series and not override:
                raise Exception(f'Type {klass} claims dtype (or dtype alias) {dtype_alias}, which is '
                                f'already assigned to {self.dtype_to_series[dtype_alias]}')
            self.dtype_to_series[dtype_alias] = klass

    def _register_db_dtype_klass(self, klass: Type['Series'], override=False):
        for db_dialect in klass.supported_db_dtype.keys():
            if db_dialect not in self.db_dtype_to_series:
                self.db_dtype_to_series[db_dialect] = {}
        for db_dialect, db_dtype in klass.supported_db_dtype.items():
            if db_dtype is None:
                continue
            if db_dtype in self.db_dtype_to_series.get(db_dialect, {}) and not override:
                raise Exception(f'Type {klass} claims db_dtype {db_dtype} for {db_dialect.value}, which is '
                                f'already assigned to dtype {self.db_dtype_to_series[db_dialect][db_dtype]}')
            self.db_dtype_to_series[db_dialect][db_dtype] = klass

    def _register_value_klass(self, value_type: Type, klass: Type['Series'], override=False):
        for vt, kt in self.value_type_to_series:
            if vt == value_type and kt != klass and not override:
                raise Exception(f'Cannot register {value_type} twice, already coupled to {klass}')
        if value_type not in klass.supported_value_types:
            raise ValueError(f'Cannot register {klass} for {value_type}. Type not supported.')
        type_tuple = value_type, klass
        self.value_type_to_series.append(type_tuple)

    def register_dtype_series(self,
                              series_type: Type['Series'],
                              value_types: List[Type],
                              override_registered_types: bool = False):
        """
        Add a Series sub-class to this registry.
        Will register the given series_type as the default type for its dtype and db_dtype
        """
        self._real_init()
        self._register_dtype_klass(series_type, override_registered_types)
        self._register_db_dtype_klass(series_type, override_registered_types)
        for value_type in value_types:
            self._register_value_klass(value_type, series_type, override_registered_types)

    def get_series_type_from_dtype(self, dtype: DtypeOrAlias) -> Type['Series']:
        """
        Given a dtype string or a dtype alias, will return the correct Series object to represent such data.
        """
        self._real_init()
        # lists and dicts are hardcoded exceptions for now.
        # TODO: make this nicer, e.g. make sure that no class can register this as an alias
        #  maybe switch to all strings, e.g. `'list[int64]'` instead of `list['int64']` ?
        if isinstance(dtype, list):
            dtype = 'list'
        elif isinstance(dtype, dict):
            dtype = 'dict'

        if dtype not in self.dtype_to_series:
            raise ValueError(f'Unknown dtype: {dtype}')
        return self.dtype_to_series[dtype]

    def get_dtype_from_db_dtype(self, db_dialect: DBDialect, db_dtype: str) -> StructuredDtype:
        """
        Given a db_dtype string, will return the full StructuredDtype.

        :return: dtype can be a simple dtype such as 'int64', or a structured dtype, such
                as {'a': 'int64', 'b': 'float64'}
        """
        self._real_init()
        if db_dtype in self.db_dtype_to_series[db_dialect]:
            return self.db_dtype_to_series[db_dialect][db_dtype].dtype
        if db_dialect == DBDialect.BIGQUERY:
            # TODO: make this nicer. clean up when we support more than two databases
            from bach.types_bq import bq_db_dtype_to_dtype
            return bq_db_dtype_to_dtype(db_dtype=db_dtype)
        raise ValueError(f'Unknown db_dtype: {db_dtype}')

    def get_all_db_dtype_to_series(self) -> Mapping[DBDialect, Mapping[Dtype, Type['Series']]]:
        """ Get all registered mappings from db_dtype to Series type"""
        self._real_init()
        return {db_dialect: copy(mapping) for db_dialect, mapping in self.db_dtype_to_series.items()}

    def value_to_dtype(self, value: Any) -> Dtype:
        """
        Given a python value, return the dtype string of the Series that's registered as the default
        for the type of value.
        For non-scalar types this will give the base Dtype, not the StructuredDtype. E.g. for an list of
            integers this will return 'list' and not ['int64']
        """
        self._real_init()
        # exception for values that are Series. Check: do we need this exception?
        from bach.series import Series
        if isinstance(value, Series):
            return value.dtype
        # iterate in reverse, the last item added that matches is used in case where multiple entries
        # match.
        for type_object, series_type in self.value_type_to_series[::-1]:
            if isinstance(value, type_object):
                return series_type.dtype
        raise ValueError(f'No dtype known for {type(value)}')

    def validate_is_dtype(self, dtype: StructuredDtype):
        """
        Validate that dtype is a valid dtype, either as a regular dtype string that is one of the registered
        dtypes, or a structured dtype that is well-formed.
        :raises ValueError: if dtype is not a valid Dtype
        """
        self._real_init()
        base_dtypes = set(series.dtype for series in self.dtype_to_series.values())
        _assert_is_valid_dtype(base_dtypes=base_dtypes, dtype=dtype)


_registry = TypeRegistry()


def _validate_dtype_of_value(dtype: StructuredDtype, value: Union['Series', AllSupportedLiteralTypes]):
    """
    Recursively validate that value has the given dtype.
    Assumes dtype is a well-formed dtype.
    """
    from bach.series import Series

    if isinstance(value, Series):
        # If value is already a series with a dtype, it should match dtype exactly.
        if dtype != value.dtype:
            raise ValueError(f'Dtype does not match dtype of series. '
                             f'Dtype: {dtype}, series.dtype: {value.dtype}')

    elif isinstance(dtype, Dtype):
        series = get_series_type_from_dtype(dtype)
        if type(value) not in series.supported_value_types:
            raise ValueError(f'Dtype does not match value. '
                             f'Dtype: {dtype}, '
                             f'supported value types: {series.supported_value_types}, '
                             f'value: {value}')
    elif isinstance(dtype, list):
        if not isinstance(value, list):
            raise ValueError(f'Dtype is a list, value is not a list. Value: {value}')
        sub_dtype = dtype[0]
        for sub_value in value:
            _validate_dtype_of_value(sub_dtype, sub_value)

    elif isinstance(dtype, dict):
        if not isinstance(value, dict):
            raise ValueError(f'Dtype is a dict, value is not a dict. Value: {value}')
        dtype_keys = set(dtype.keys())
        value_keys = set(value.keys())
        if dtype_keys != value_keys:
            raise ValueError(f'Dtype keys do not match value keys. '
                             f'Dtype keys: {dtype_keys}, value keys: {value_keys}')
        for key, sub_dtype in dtype.items():
            sub_value = value[key]
            _validate_dtype_of_value(sub_dtype, sub_value)
    else:
        raise ValueError(f'Dtype not supported. Dtype: {dtype}')


def _assert_is_valid_dtype(base_dtypes: Set[Dtype], dtype: StructuredDtype):
    """
    Validate that dtype is a valid dtype, either a regular dtype or a recursively defined structural dtype.
    :raise ValueError: if dtype is not a valid Dtype
    """
    if isinstance(dtype, Dtype):
        if dtype not in base_dtypes:
            raise ValueError(f'Unknown dtype string: {dtype}')
    elif isinstance(dtype, list):
        if len(dtype) != 1:
            raise ValueError(f'Expected list of length one, got: {dtype}')
        sub_dtype = dtype[0]
        _assert_is_valid_dtype(base_dtypes=base_dtypes, dtype=sub_dtype)
    elif isinstance(dtype, dict):
        non_string_keys = sorted(key for key in dtype.keys() if not isinstance(key, str))
        if non_string_keys:
            raise ValueError(f'Expected all keys to be strings. '
                             f'Found non string keys: {non_string_keys}')
        for sub_dtype in dtype.values():
            _assert_is_valid_dtype(base_dtypes=base_dtypes, dtype=sub_dtype)
    else:
        raise ValueError(f'Unknown dtype: {dtype}')



"""
Copyright 2021 Objectiv B.V.
"""
import math
from abc import ABC
from typing import cast, Union, TYPE_CHECKING, Optional, List, Tuple

import numpy
from sqlalchemy.engine import Dialect

from bach import DataFrameOrSeries
from bach.series import Series
from bach.expression import Expression, AggregateFunctionExpression
from bach.series.series import WrappedPartition
from bach.types import StructuredDtype, AllSupportedLiteralTypes
from sql_models.constants import DBDialect
from sql_models.util import is_postgres, is_bigquery, DatabaseNotSupportedException, is_athena

if TYPE_CHECKING:
    from bach.series import SeriesBoolean, SeriesNumericInterval


class SeriesAbstractNumeric(Series, ABC):
    """
    Abstract Series class that represents the base numeric types and its specific operations.

    **Operations**

    All common arithmetic operations are supported, as well as common aggregation functions, and statistical
    functions.

    Supported arithmetic operations:
    - add (+), subtract (-)
    - multiply (*), divide (/), floordiv (//), mod (%)
    - lshift (<<) and rshift(>>) for Integer types

    """
    def _arithmetic_operation(self, other, operation, fmt_str,
                              other_dtypes=('int64', 'float64'), dtype=None):
        return super()._arithmetic_operation(other, operation, fmt_str, other_dtypes, dtype)

    def _comparator_operation(self, other, comparator, other_dtypes=('int64', 'float64'),
                              strict_other_dtypes=tuple()) -> 'SeriesBoolean':
        return super()._comparator_operation(other, comparator, other_dtypes, strict_other_dtypes)

    def exp(self) -> 'SeriesAbstractNumeric':
        """
        Get exp of value
        """
        return self.copy_override(
            expression=Expression.construct(f'exp({{}})', self)
        )

    def round(self, decimals: int = 0) -> 'SeriesAbstractNumeric':
        """
        Round the value of this series to the given amount of decimals.

        :param decimals: The amount of decimals to round to
        """
        if is_athena(self.engine):
            return self.copy_override(
                expression=Expression.construct(f'round({{}}, {decimals})', self)
            )
        return self.copy_override(
            expression=Expression.construct(f'round(cast({{}} as numeric), {decimals})', self)
        )

    def cut(self, bins: int, right: bool = True) -> 'SeriesNumericInterval':
        """
        Segments values into bins.

        :param bins: The amount of bins to segment data into
        :param right: If true (by default), each bin will include the rightmost edge. (e.g (x,y]).
        """
        from bach.operations.cut import CutOperation
        return CutOperation(series=self, bins=bins, right=right)()

    def qcut(self, q: Union[int, List[float]]) -> 'SeriesNumericInterval':
        """
        Segments values into equal-sized buckets based on rank or sample quantiles.

        :param q: Number of quantiles or list of quantiles to consider.

        :return: series containing each quantile range/interval per value. Original series is set as index.
        """
        from bach.operations.cut import QCutOperation
        return QCutOperation(series=self, q=q)()

    def _ddof_unsupported(self, ddof: Optional[int]):
        if ddof is not None and ddof != 1:
            raise NotImplementedError("ddof != 1 currently not implemented")

    # def kurt(self, partition: WrappedPartition = None, skipna: bool = True):
    #     return self.kurtosis(partition, skipna)
    #
    # def kurtosis(self, partition: WrappedPartition = None, skipna: bool = True):
    #     raise NotImplementedError("kurtosis currently not implemented")
    #
    # def mad(self, partition: WrappedPartition = None, skipna: bool = True):
    #     raise NotImplementedError("mad currently not implemented")
    #
    # def prod(self, partition: WrappedPartition = None, skipna: bool = True):
    #     return self.product(partition, skipna)
    #
    # def product(self, partition: WrappedPartition = None, skipna: bool = True):
    #     raise NotImplementedError("prod currently not implemented")
    #
    # def skew(self, partition: WrappedPartition = None, skipna: bool = True):
    #     raise NotImplementedError("skew currently not implemented")

    def sem(self, partition: WrappedPartition = None, skipna: bool = True, ddof: int = None, **kwargs):
        """
        Get the unbiased standard error of the mean.
        Normalized by N-1 by default.

        :param partition: The partition or window to apply
        :param skipna: Exclude NA/NULL values
        :param ddof: Delta degrees of freedom. he divisor used in calculations is N - ddof,
            where N represents the number of elements
        """
        self._ddof_unsupported(ddof)
        return self._derived_agg_func(
            partition,
            AggregateFunctionExpression.construct(f'{{}}/sqrt({{}})',
                                                  self.std(partition, skipna=skipna, ddof=ddof),
                                                  self.count(partition, skipna=skipna)),
            skipna=skipna
        )

    def std(self, partition: WrappedPartition = None, skipna: bool = True, ddof: int = None, **kwargs):
        """
        Get the standard deviation of the input values
        Normalized by N-1 by default.

        :param partition: The partition or window to apply
        :param skipna: Exclude NA/NULL values
        :param ddof: Delta degrees of freedom. The divisor used in calculations is N - ddof,
            where N represents the number of elements
        """
        if ddof is not None and ddof not in (0, 1):
            raise NotImplementedError(f"ddof == {ddof} currently not implemented")

        if ddof == 0:
            return self._derived_agg_func(partition, 'stddev_pop', skipna=skipna)
        return self._derived_agg_func(partition, 'stddev_samp', skipna=skipna)

    def sum(self, partition: WrappedPartition = None, skipna: bool = True, min_count: int = None, **kwargs):
        """
        Get the sum of the input values.

        :param partition: The partition or window to apply
        :param skipna: Exclude NA/NULL values
        :param min_count: This minimum amount of values (not NULL) to be present before returning a result.
        """
        return self._derived_agg_func(partition, 'sum', skipna=skipna, min_count=min_count)

    def mean(self, partition: WrappedPartition = None, skipna: bool = True, **kwargs) -> 'SeriesFloat64':
        """
        Get the mean/average of the input values.

        :param partition: The partition or window to apply
        :param skipna: Exclude NA/NULL values
        """
        return cast(
            'SeriesFloat64',  # for the mypies
            self._derived_agg_func(partition, 'avg', 'double precision', skipna=skipna),
        )

    def quantile(
        self, partition: WrappedPartition = None, q: Union[float, List[float]] = 0.5, **kwargs
    ) -> 'SeriesFloat64':
        """
        When q is a float or len(q) == 1, the resultant series index will remain
        In case multiple quantiles are calculated, the resultant series index will have all calculated
        quantiles as index values.

        :param partition: The partition or window to apply
        :param q: A quantile or list of quantiles to be calculated
        """
        result = self.to_frame().quantile(q=q, partition=partition)[f'{self.name}_quantile']
        result = result.copy_override(name=self.name)

        if not isinstance(q, list):
            result = result.reset_index(level='quantile', drop=True)
        return cast('SeriesFloat64', result)

    def var(self, partition: WrappedPartition = None, skipna: bool = True, ddof: int = None, **kwargs):
        """
        Get the sample variance of the input values (square of the sample standard deviation)
        Normalized by N-1 by default.

        :param partition: The partition or window to apply
        :param skipna: Exclude NA/NULL values
        :param ddof: Delta degrees of freedom. he divisor used in calculations is N - ddof,
            where N represents the number of elements
        """
        self._ddof_unsupported(ddof)
        return self._derived_agg_func(partition, 'var_samp', skipna=skipna)

    def scale(self, with_mean: bool = True, with_std: bool = True) -> 'Series':
        """
        Standardizes series based on mean and population standard deviation.

        :param with_mean: if true, each value will be centered before scaling
        :param with_std: if true, each value will be scaled to unit variance

        :return: Series
        """
        return self.to_frame().scale(with_mean, with_std)[self.name]

    def minmax_scale(self, feature_range: Tuple[int, int] = (0, 1)) -> 'Series':
        """
        Scales series based on a given range.

        :param feature_range: ``tuple(min, max)`` desired range to use for scaling.
        :return: Series
        """
        return self.to_frame().minmax_scale(feature_range)[self.name]


class SeriesInt64(SeriesAbstractNumeric):
    """
    Series type for integer data.

    **Operations and Functions**

    All common arithmetic operations are supported, as well as common aggregation functions, and statistical
    functions.

    Supported arithmetic operations:
    - add (+), subtract (-)
    - multiply (*), divide (/), floordiv (//), mod (%)
    - lshift (<<) and rshift(>>)

    See the base class :class:`SeriesAbstractNumeric` for an overview of the supported aggregation and
    statistical functions.


    **Database support and types**

    * Postgres: utilizes the 'bigint' database type.
    * Athena: utilizes the 'bigint' database type.
    * BigQuery: utilizes the 'INT64' database type.
    """
    dtype = 'int64'
    dtype_aliases = ('integer', 'bigint', 'i8', int, numpy.int64, 'int32')
    supported_db_dtype = {
        DBDialect.POSTGRES: 'bigint',
        DBDialect.ATHENA: 'bigint',
        DBDialect.BIGQUERY: 'INT64'
    }
    supported_value_types = (int, numpy.int64, numpy.int32)
    supported_source_dtypes = ('float64', 'bool', 'string')

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        # We override the parent class here because integers are really common, and we don't strictly need
        # to cast all of them for BigQuery. Not casting integer literals greatly improves the readability of
        # the generated SQL.

        if is_postgres(dialect) or is_athena(dialect):
            # A stringified integer is a valid integer or bigint literal, depending on the size. We want to
            # consistently get bigints, so always cast the result
            # PG: See the section on numeric constants in the Postgres documentation
            # https://www.postgresql.org/docs/14/sql-syntax-lexical.html#SQL-SYNTAX-CONSTANTS
            # Athena: experimentally established that the integer typing mimicks Postgres's behaviour
            return super().supported_literal_to_expression(dialect=dialect, literal=literal)
        if is_bigquery(dialect):
            # BigQuery has only one integer type, so there is no confusion between 32-bit and 64-bit integers
            # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#integer_type
            # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#integer_literals
            # Additionally, we don't even need to cast NULL, because on BigQuery that's an INT64 by default.
            return literal
        raise DatabaseNotSupportedException(dialect)

    @classmethod
    def supported_value_to_literal(
            cls,
            dialect: Dialect,
            value: int,
            dtype: StructuredDtype
    ) -> Expression:
        return Expression.raw(str(value))

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        if source_dtype == 'int64':
            return expression
        if source_dtype not in cls.supported_source_dtypes:
            raise ValueError(f'cannot convert {source_dtype} to int64')
        return Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)

    def _arithmetic_operation(self, other, operation, fmt_str, other_dtypes=('int64', 'float64'), dtype=None):
        # Override this method, because we need to return a float if we interact with one.
        type_mapping = dtype if dtype else {
            'int64': 'int64',
            'float64': 'float64'
        }
        return super()._arithmetic_operation(other, operation, fmt_str, other_dtypes, type_mapping)

    def __truediv__(self, other) -> 'Series':
        return self.astype('float64') / other

    def __rshift__(self, other):
        if is_postgres(self.engine):
            # Postgres expects the argument to a bitshift to be a regular int, not bigint.
            return self._arithmetic_operation(other, 'rshift', '({}) >> cast({} as int)',
                                              other_dtypes=tuple(['int64']))
        if is_athena(self.engine):
            # Latest version of prestodb supports bitwise_logical_shift_right() [1], but the version that
            # Athena uses, does not [2]. So we do some hackish thing here
            # [1] https://prestodb.io/docs/current/functions/bitwise.html
            # [2] https://prestodb.io/docs/0.217/functions/bitwise.html
            return self._arithmetic_operation(other, 'rshift', 'cast(({}) / power(2, {}) as bigint)',
                                              other_dtypes=tuple(['int64']))
        return self._arithmetic_operation(other, 'rshift', '({}) >> ({})', other_dtypes=tuple(['int64']))

    def __lshift__(self, other):
        if is_postgres(self.engine):
            # Postgres expects the argument to a bitshift to be a regular int, not bigint.
            return self._arithmetic_operation(other, 'lshift', '({}) << cast({} as int)',
                                              other_dtypes=tuple(['int64']))
        if is_athena(self.engine):
            # hack around missing shift operator
            return self._arithmetic_operation(other, 'lshift', 'cast(({}) * power(2, {}) as bigint)',
                                              other_dtypes=tuple(['int64']))

        return self._arithmetic_operation(other, 'lshift', '({}) << ({})', other_dtypes=tuple(['int64']))

    def sum(self, partition: WrappedPartition = None, skipna: bool = True, min_count: int = None, **kwargs):
        # sum() has the tendency to return float on bigint arguments. Cast it back.
        series = super().sum(partition, skipna, min_count)
        return series.copy_override(
            expression=Expression.construct('cast({} as bigint)', series.expression))

    def round(self, decimals: int = 0) -> 'SeriesAbstractNumeric':
        # round() should not affect int dtype series.
        return self


class SeriesFloat64(SeriesAbstractNumeric):
    """
    Series type for float data.

    **Operations and Functions**

    All common arithmetic operations are supported, as well as common aggregation functions, and statistical
    functions.

    Supported arithmetic operations:
    - add (+), subtract (-)
    - multiply (*), divide (/), floordiv (//), mod (%)

    See the base class :class:`SeriesAbstractNumeric` for an overview of the supported aggregation and
    statistical functions.

    **Database support and types**

    * Postgres: utilizes the 'double precision' database type.
    * Athena: utilizes the 'double' database type.
    * BigQuery: utilizes the 'FLOAT64' database type.
    """
    dtype = 'float64'
    dtype_aliases = ('float', 'double', 'f8', float, numpy.float64, 'double precision')
    supported_db_dtype = {
        DBDialect.POSTGRES: 'double precision',
        DBDialect.ATHENA: 'double',
        DBDialect.BIGQUERY: 'FLOAT64'
    }
    supported_value_types = (float, numpy.float64)
    supported_source_dtypes = ('int64', 'string')

    # Notes for supported_value_to_literal() and supported_literal_to_expression():
    #
    # ### Postgres ###
    # Postgres will automatically parse any number with a decimal point as a number of type `numeric`,
    # which could be casted to float. However we specify the value always as a string, as there are some
    # values that cannot be expressed as a numeric literal directly (NaN, infinity, and -infinity), and
    # a value that cannot be represented as numeric (-0.0).
    # See the sections on numeric constants, and on fLoating-point types in the Postgres documentation
    # https://www.postgresql.org/docs/14/sql-syntax-lexical.html#SQL-SYNTAX-CONSTANTS
    # https://www.postgresql.org/docs/14/datatype-numeric.html#DATATYPE-FLOAT
    #
    # ### BigQuery ###
    # BigQuery parses floating point literals as a float, no casts needed. However the special values (NaN,
    # infinity, and -infinity) can only be specified as strings that are then cast. For simplicity, we
    # always use a string literal that we then cast to float, just as we do for Postgres.
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#floating_point_types
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#floating_point_literals

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: Union[float, numpy.float64],
        dtype: StructuredDtype
    ) -> Expression:
        if is_athena(dialect) and not math.isfinite(value):
            # see https://prestodb.io/docs/0.217/functions/math.html, under 'Floating Point functions'
            if math.isnan(value):
                return Expression.raw('nan()')
            if math.isinf(value):
                if value > 0:
                    return Expression.raw('infinity()')
                else:
                    return Expression.raw('-infinity()')
        return Expression.string_value(str(value))

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        if source_dtype == 'float64':
            return expression
        if source_dtype not in cls.supported_source_dtypes:
            raise ValueError(f'cannot convert {source_dtype} to float64')
        return Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)

    @classmethod
    def random(cls, base: DataFrameOrSeries) -> 'SeriesFloat64':
        """
        Create a new Series object with an expression, that will evaluate to a random float in the range
        [0, 1) for each row.

        :param base: DataFrame or Series from which the newly created Series' engine, base_node and index
            parameters are copied.

        .. warning::
            The returned Series has a non-deterministic expression, it will give a different result each
            time it is evaluated by the database.

        The non-deterministic expression can have some unexpected consequences. Consider the following code:

        .. code-block:: python

            df['x'] = SeriesFloat64.random(df)
            df['y'] = df['x']
            df['different'] = df['y'] != df['x']

        The df['different'] series will be True for (almost) all rows, because the second statement copies
        the unevaluated expression, not the result of the expression. So at evaluation time the expression
        will be evaluated twice for each row, for the 'x' column and the 'y' column, giving different
        results both times. One way to work around this is to materialize the dataframe in its current state
        (using materialize()), before adding any columns that reference a column that's created with
        this function.
        """
        if is_postgres(base.engine) or is_athena(base.engine):
            expr_str = 'random()'
        elif is_bigquery(base.engine):
            expr_str = 'RAND()'
        else:
            raise DatabaseNotSupportedException(base.engine)
        return cls.get_class_instance(
            engine=base.engine,
            base_node=base.base_node,
            index=base.index,
            name='random',
            expression=Expression.construct(expr_str),
            group_by=None,
            order_by=[],
            instance_dtype=cls.dtype
        )



"""
Copyright 2021 Objectiv B.V.
"""
import datetime
from abc import ABC
from enum import Enum
from typing import Union, cast, List, Tuple, Optional, Any, Type

import numpy
import pandas
from sqlalchemy.engine import Dialect

from bach import DataFrame
from bach.series import Series, SeriesString, SeriesBoolean, SeriesFloat64, SeriesInt64, SeriesAbstractNumeric
from bach.expression import Expression, join_expressions, StringValueToken
from bach.series.series import WrappedPartition, ToPandasInfo
from bach.series.utils.datetime_formats import parse_c_standard_code_to_postgres_code, \
    parse_c_code_to_bigquery_code, warn_non_supported_format_codes, parse_c_code_to_athena_expression
from bach.types import DtypeOrAlias, StructuredDtype, value_to_series_type, get_series_type_from_dtype
from sql_models.constants import DBDialect
from sql_models.util import is_postgres, is_bigquery, DatabaseNotSupportedException, is_athena


class DatePart(str, Enum):
    DAY = 'days'
    HOUR = 'hours'
    MINUTE = 'minutes'
    SECOND = 'seconds'
    MILLISECOND = 'milliseconds'
    MICROSECOND = 'microseconds'


# conversions for date parts to seconds
# when adjusting intervals, 30-day time periods are represented as months
# BigQuery seems to follow Postgres threshold
# https://www.postgresql.org/docs/current/functions-datetime.html#:~:text=justify_days%20(%20interval%20)%20%E2%86%92%20interval,mon%205%20days
# For example 395 days is equal to 1 year, 1 month and 5 days.
_TOTAL_SECONDS_PER_DATE_PART = {
    DatePart.DAY:  24 * 60 * 60,
    DatePart.HOUR: 60 * 60,
    DatePart.MINUTE: 60,
    DatePart.SECOND: 1,
    DatePart.MILLISECOND: 1e-3,
    DatePart.MICROSECOND: 1e-6,
}

AllSupportedDateTimeTypes = Union[
    datetime.datetime,
    numpy.datetime64,
    datetime.date,
    str,
    None,
    datetime.timedelta,
    numpy.timedelta64,
]


class DateTimeOperation:
    def __init__(self, series: 'SeriesAbstractDateTime'):
        self._series = series

    def strftime(self, format_str: str) -> SeriesString:
        """
        Allow formatting of this Series (to a string type).

        :param format_str: The format to apply to the date/time column. See Supported Format Codes below for
            more information on what is supported.

        :returns: a SeriesString containing the formatted date.

        **Supported Format Codes**

        A subset of the C standard format codes is supported. See the python documentation for the
        code semantics: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

        The subset of codes that are supported across all databases is:
            `%A`, `%B`, `%F`, `%H`, `%I`, `%M`, `%R`, `%S`, `%T`, `%V`, %Y`, `%a`, `%b`, `%d`, `%j`, `%m`,
            `%y`, `%%`

        Additionally one specific combination is supported: `%S.%f`

        **Example**

        .. code-block:: python

            df['year'] = df.some_date_series.dt.strftime('%Y')  # return year
            df['date'] = df.some_date_series.dt.strftime('%Y%m%d')  # return date
        """
        engine = self._series.engine
        series = self._series

        # Warn user if he is using non-supported format codes
        warn_non_supported_format_codes(format_str)

        if is_postgres(engine):
            parsed_format_str = parse_c_standard_code_to_postgres_code(format_str)
            expression = Expression.construct(
                'to_char({}, {})', series, Expression.string_value(parsed_format_str)
            )
        elif is_athena(engine):
            if isinstance(series, SeriesTime):
                # for time series, we should convert the float values into a timestamp,
                # otherwise date_format won't work
                series = SeriesTimestamp.from_total_seconds(series.copy_override_type(SeriesFloat64))

            expression = parse_c_code_to_athena_expression(series.expression, format_str)
        elif is_bigquery(engine):
            if isinstance(series, SeriesTime):
                # BigQuery provides a different function for formatting time, parsing a time to timestamp
                # is not an option as the casting is not supported.
                fmt_str = 'format_time({}, {})'
            else:
                # BigQuery will ignore some formatting codes when applied to a date.
                # For example, rather than returning '00' as value for '%H' (hour) it will return '%H'.
                # To avoid this problem, we convert to timestamp.
                series = cast(SeriesTimestamp, series.astype('timestamp'))
                fmt_str = 'format_date({}, {})'

            parsed_format_str = parse_c_code_to_bigquery_code(format_str)
            expression = Expression.construct(fmt_str, Expression.string_value(parsed_format_str), series)
        else:
            raise DatabaseNotSupportedException(engine)

        str_series = series.copy_override_type(SeriesString).copy_override(expression=expression)
        return str_series

    def date_trunc(self, date_part: str) -> Series:
        """
        Truncates date value based on a specified date part.
        The value is always rounded to the beginning of date_part.

        This operation can be applied only on SeriesDate or SeriesTimestamp.

        :param date_part: Allowed values are 'second', 'minute',
            'hour', 'day', 'week', 'month', 'quarter', and 'year'.

        .. code-block:: python

            # return the date corresponding to the Monday of that week
            df['week'] = df.some_date_or_timestamp_series.dt.date_trunc('week')
            # return the first day of the quarter
            df['quarter'] = df.some_date_or_timestamp_series.dt.date_trunc('quarter')

        :returns: the truncated timestamp value with a granularity of date_part.

        """

        available_formats = ['second', 'minute', 'hour', 'day', 'week',
                             'month', 'quarter', 'year']
        if date_part not in available_formats:
            raise ValueError(f'{date_part} format is not available.')

        if not (isinstance(self._series, SeriesDate) or
                isinstance(self._series, SeriesTimestamp)):
            raise ValueError(f'{type(self._series)} type is not supported.')

        engine = self._series.engine

        # Truncating Postgres date type will include timezone in final value
        # therefore we should cast it into TIMESTAMP WITHOUT TIMEZONE
        _td_series = self._series.astype(SeriesTimestamp.dtype)
        if is_postgres(engine) or is_athena(engine):
            expression = Expression.construct(
                'date_trunc({}, {})',
                Expression.string_value(date_part),
                _td_series,
            )
        elif is_bigquery(engine):
            if date_part == 'week':
                date_part = 'week(monday)'
            expression = Expression.construct(
                'timestamp_trunc({}, {})',
                _td_series,
                Expression.raw(date_part),
            )
        else:
            raise DatabaseNotSupportedException(engine)

        return _td_series.copy_override(expression=expression)


class TimedeltaOperation(DateTimeOperation):
    def _get_conversion_df(self) -> 'DataFrame':
        """
        generates a dataframe containing the amounts of seconds a supported date part has.
        """
        from bach import DataFrame
        conversion_df = pandas.DataFrame(
            data=[
                {
                    self._format_converted_series_name(dp): ts
                    for dp, ts in _TOTAL_SECONDS_PER_DATE_PART.items()
                },
            ]
        )
        convert_df = DataFrame.from_pandas(df=conversion_df, engine=self._series.engine)
        return convert_df.reset_index(drop=True)

    @staticmethod
    def _format_converted_series_name(date_part: DatePart) -> str:
        return f'_SECONDS_IN_{date_part.name}'.lower()

    @property
    def components(self) -> DataFrame:
        """
        :returns: a DataFrame containing all date parts from the timedelta.
        """
        df = self.total_seconds.to_frame()
        df = df.merge(self._get_conversion_df(), how='cross')

        # justifies total seconds into the units of each date component
        # after adjustment, it converts it back into seconds
        for date_part in DatePart:
            converted_series_name = self._format_converted_series_name(DatePart(date_part))
            df[f'ts_{date_part}'] = df['total_seconds'] // df[converted_series_name]
            df[f'ts_{date_part}'] *= df[converted_series_name]

        # materialize to avoid complex subquery
        df = df.materialize(node_name='justified_date_components')

        components_series_names = []
        prev_ts = ''

        # extract actual date component from justified seconds
        # by getting the difference between current and previous components
        # this helps on normalizing negative time deltas and have only negative values
        # in days.
        for date_part in DatePart:
            converted_series_name = self._format_converted_series_name(DatePart(date_part))

            component_name = f'{date_part}'
            current_ts = f'ts_{date_part}'

            if not prev_ts:
                df[component_name] = df[current_ts] / df[converted_series_name]
            else:
                df[component_name] = (df[current_ts] - df[prev_ts]) / df[converted_series_name]

            df[component_name] = cast(SeriesFloat64, df[component_name]).round(decimals=0)

            components_series_names.append(component_name)
            prev_ts = current_ts

        return df[components_series_names].astype('int64')

    @property
    def days(self) -> SeriesInt64:
        """
        converts total seconds into days and returns only the integral part of the result
        """
        day_series = self.total_seconds // _TOTAL_SECONDS_PER_DATE_PART[DatePart.DAY]

        day_series = day_series.astype('int64')
        return (
            day_series
            .copy_override_type(SeriesInt64)
            .copy_override(name='days')
        )

    @property
    def seconds(self) -> SeriesInt64:
        """
        removes days from total seconds (self.total_seconds % _SECONDS_IN_DAY)
        and returns only the integral part of the result
        """
        seconds_series = (self.total_seconds % _TOTAL_SECONDS_PER_DATE_PART[DatePart.DAY]) // 1

        seconds_series = seconds_series.astype('int64')
        return (
            seconds_series
            .copy_override_type(SeriesInt64)
            .copy_override(name='seconds')
        )

    @property
    def microseconds(self) -> SeriesInt64:
        """
        considers only the fractional part of the total seconds and converts it into microseconds
        """
        microseconds_series = (
            (self.total_seconds % 1) / _TOTAL_SECONDS_PER_DATE_PART[DatePart.MICROSECOND]
        )

        microseconds_series = microseconds_series.astype('int64')
        return (
            microseconds_series
            .copy_override_type(SeriesInt64)
            .copy_override(name='microseconds')
        )

    @property
    def total_seconds(self) -> SeriesFloat64:
        """
        returns the total amount of seconds in the interval

        .. note::
            Since Athena does not support interval types, series values are actually the total seconds.
        """
        expression = self._series.expression
        if is_postgres(self._series.engine):
            # extract(epoch from source) returns the total number of seconds in the interval
            expression = Expression.construct(f'extract(epoch from {{}})', self._series)
        elif is_bigquery(self._series.engine):
            # bq cannot extract epoch from interval
            expression = Expression.construct(
                (
                    f"UNIX_MICROS(CAST('1970-01-01' AS TIMESTAMP) + {{}}) "
                    f"* {_TOTAL_SECONDS_PER_DATE_PART[DatePart.MICROSECOND]}"
                ),
                self._series,
            )

        return (
            self._series
            .copy_override_type(SeriesFloat64)
            .copy_override(name='total_seconds', expression=expression)
        )


class SeriesAbstractDateTime(Series, ABC):
    """
    A Series that represents the generic date/time type and its specific operations. Selected arithmetic
    operations are accepted using the usual operators.

    **Date/Time Operations**

    On any of the subtypes, you can access date operations through the `dt` accessor.
    """

    # list of formats used for parsing string literals to date object supported by Series
    _VALID_DATE_TIME_FORMATS = [
        '%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M', '%Y-%m-%d'
    ]

    # format used for all literals when casting them to this Series
    _FINAL_DATE_TIME_FORMAT = ''

    @property
    def dt(self) -> DateTimeOperation:
        """
        Get access to date operations.

        .. autoclass:: bach.series.series_datetime.DateTimeOperation
            :members:

        """
        return DateTimeOperation(self)

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        if source_dtype == cls.dtype:
            return expression

        if source_dtype not in cls.supported_source_dtypes:
            raise ValueError(f'cannot convert {source_dtype} to {cls.dtype}')

        if source_dtype == 'string':
            if expression.is_constant:
                return cls._constant_expression_to_dtype_expression(dialect, expression)
            if is_athena(dialect) and cls.dtype == 'time':
                # casting varchar to time is a bit tricky, since the string value might not contain
                # fractional part of the seconds, this breaks date_parse function
                expression = Expression.construct(
                    "COALESCE(try(date_parse({}, '%H:%i:%S.%f')), try(date_parse({}, '%H:%i:%S')))",
                    expression,
                    expression,
                )
                expression = Expression.construct('to_unixtime({})', expression)

        return Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)

    def astype(self, dtype: Union[str, Type]) -> Series:
        if get_series_type_from_dtype(dtype) is not SeriesString:
            return super().astype(dtype=dtype)

        return self.dt.strftime(self._FINAL_DATE_TIME_FORMAT)

    def _comparator_operation(self, other, comparator,
                              other_dtypes=('timestamp', 'date', 'time', 'string'),
                              strict_other_dtypes=tuple()) -> 'SeriesBoolean':
        return super()._comparator_operation(
            other, comparator, other_dtypes, strict_other_dtypes=tuple(self.dtype)
        )

    @classmethod
    def _cast_and_round_to_date(cls, series: 'SeriesDate') -> 'Series':
        # Most of engines return timestamp in all cases were we expect date
        # Make sure we cast properly, and round similar to python datetime: add 12 hours and cast to date
        td_12_hours = datetime.timedelta(seconds=3600 * 12)
        series_12_hours = SeriesTimedelta.from_value(base=series, value=td_12_hours, name='tmp')
        expr_12_hours = series_12_hours.expression

        expr = "cast({} + {} as date)"

        # time intervals are not supported in athena, we need to use unix time instead
        if is_athena(series.engine):
            expr = "cast(from_unixtime(to_unixtime({}) + {}) as date)"

        return series.copy_override(
            expression=Expression.construct(expr, series, expr_12_hours)
        )

    @classmethod
    def _get_date_format_for_literal_value(cls, value: str) -> str:
        """
        Returns the correct date format for a string literal. If value has a non-supported format,
        an exception will be raised.
        """
        for str_format in cls._VALID_DATE_TIME_FORMATS:
            try:
                datetime.datetime.strptime(value, str_format)
            except ValueError:
                continue
            else:
                return str_format

        raise ValueError(
            f'Not a valid string literal: {value} for {cls.dtype} dtype.'
            f'Supported formats: {cls._VALID_DATE_TIME_FORMATS}'
        )

    @classmethod
    def _constant_expression_to_dtype_expression(cls, dialect, expression) -> Expression:
        """
        Parses constant expression to respective dtype expression. Child classes must be
        in charge of validating if string literals have the expected date format.
        """
        # check if string literal has correct format
        literal_expr_tokens = []
        for token in expression.data:
            if not isinstance(token, StringValueToken):
                literal_expr_tokens.append(token)
                continue

            val = token.value if token.value != 'NULL' else None
            val_to_lit_expr = cls.supported_value_to_literal(dialect, val, cls.dtype)
            literal_expr_tokens.append(val_to_lit_expr)
        return cls.supported_literal_to_expression(dialect, Expression(literal_expr_tokens))

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: AllSupportedDateTimeTypes,
        dtype: StructuredDtype
    ) -> Expression:
        if value is None:
            return Expression.raw('NULL')

        dt_value: Union[datetime.datetime, datetime.date, datetime.time, None] = None

        if isinstance(value, str):
            # get the right date format of the string
            str_format = cls._get_date_format_for_literal_value(value)
            dt_value = datetime.datetime.strptime(value, str_format)

        if isinstance(value, numpy.datetime64):
            if numpy.isnat(value):
                return Expression.raw('NULL')

            dt_value = _convert_numpy_datetime_to_utc_datetime(value)

        if isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
            dt_value = value

        if not dt_value or not isinstance(value, cls.supported_value_types):
            raise ValueError(f'Not a valid {cls.dtype} literal: {value}')

        str_value = dt_value.strftime(cls._FINAL_DATE_TIME_FORMAT)
        return Expression.string_value(str_value)


def dt_strip_timezone(value: Optional[datetime.datetime]) -> Optional[datetime.datetime]:
    if value is None:
        return None
    return value.replace(tzinfo=None)


class SeriesTimestamp(SeriesAbstractDateTime):
    """
    A Series that represents the timestamp/datetime type and its specific operations.

    Timestamps are assumed to be in UTC, or without a timezone, both cases are treated the same.
    These timestamps have a microsecond precision at best, in contrast to numpy's datetime64 which supports
    up to attoseconds precision.

    **Database support and types**

    * Postgres: utilizes the 'timestamp without time zone' database type.
    * BigQuery: utilizes the 'TIMESTAMP' database type.
    """
    dtype = 'timestamp'
    dtype_aliases = ('datetime64', 'datetime64[ns]', numpy.datetime64)
    supported_db_dtype = {
        DBDialect.POSTGRES: 'timestamp without time zone',
        DBDialect.BIGQUERY: 'TIMESTAMP',
        DBDialect.ATHENA: 'timestamp',
    }
    supported_value_types = (datetime.datetime, numpy.datetime64, datetime.date, str)

    supported_source_dtypes = ('string', 'date',)
    _FINAL_DATE_TIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        if not is_athena(dialect):
            return super().supported_literal_to_expression(dialect, literal)

        # Athena will raise exceptions if literal has incorrect timestamp format
        return Expression.construct("date_parse({}, '%Y-%m-%d %H:%i:%S.%f')", literal)

    def to_pandas_info(self) -> Optional['ToPandasInfo']:
        if is_postgres(self.engine) or is_athena(self.engine):
            return ToPandasInfo('datetime64[ns]', None)

        if is_bigquery(self.engine):
            return ToPandasInfo('datetime64[ns, UTC]', dt_strip_timezone)
        return None

    def __add__(self, other) -> 'Series':
        _self = self.copy()
        fmt_str = '({}) + ({})'
        if is_athena(self.engine):
            # convert timestamp to unixtime and cast result back to timestamp
            _self = self.copy_override(
                expression=Expression.construct('to_unixtime({})', self)
            )
            fmt_str = f'from_unixtime({fmt_str})'

        return _self._arithmetic_operation(other, 'add', fmt_str, other_dtypes=tuple(['timedelta']))

    def __sub__(self, other) -> 'Series':
        type_mapping = {
            'timedelta': 'timestamp',
            'timestamp': 'timedelta'
        }

        _self = self.copy()
        _other = other.copy()

        fmt_str = '({}) - ({})'
        if is_athena(self.engine):
            other_series_type = value_to_series_type(other)
            # for athena, we represent timedeltas with floats, therefore is required to cast the operands
            # based on the case been evaluated.

            # timestamp - timestamp will result into a timedelta, so final result must be in terms of unixtime
            if other_series_type.dtype == SeriesTimestamp.dtype:
                fmt_str = 'round(to_unixtime({}) - to_unixtime({}), 3)'
            else:
                # timestamp - timedelta will result into a timestamp,
                # then we need to cast the unixtime to timestamp
                fmt_str = 'from_unixtime(to_unixtime({}) - {})'

        return _self._arithmetic_operation(
            _other,
            operation='sub',
            fmt_str=fmt_str,
            other_dtypes=tuple(type_mapping.keys()),
            dtype=type_mapping,
        )

    @classmethod
    def from_total_seconds(cls, total_seconds: SeriesAbstractNumeric) -> 'SeriesTimestamp':
        """
        Converts a numerical series representing total seconds (epoch) to timestamp series.

        returns a SeriesTimestamp
        """
        engine = total_seconds.engine
        if is_athena(engine):
            expression = Expression.construct('from_unixtime({})', total_seconds)

        elif is_postgres(engine):
            expression = Expression.construct('to_timestamp({})', total_seconds)
        elif is_bigquery(engine):
            # convert total seconds into microseconds
            # since TIMESTAMP_SECONDS accepts only integers, therefore
            # microseconds will be lost due to rounding
            total_microseconds_series = (
                    total_seconds / _TOTAL_SECONDS_PER_DATE_PART[DatePart.MICROSECOND]
            )
            expression = Expression.construct(
                'TIMESTAMP_MICROS({})', total_microseconds_series.astype('int64')
            )
        else:
            raise DatabaseNotSupportedException(engine)

        return total_seconds.copy_override(expression=expression).copy_override_type(SeriesTimestamp)


class SeriesDate(SeriesAbstractDateTime):
    """
    A Series that represents the date type and its specific operations

    **Database support and types**

    * Postgres: utilizes the 'date' database type.
    * BigQuery: utilizes the 'DATE' database type.
    """
    dtype = 'date'
    dtype_aliases: Tuple[DtypeOrAlias, ...] = tuple()
    supported_db_dtype = {
        DBDialect.POSTGRES: 'date',
        DBDialect.BIGQUERY: 'DATE',
        DBDialect.ATHENA: 'date',
    }
    supported_value_types = (datetime.datetime, numpy.datetime64, datetime.date, str)

    supported_source_dtypes = ('string', 'timestamp',)
    _FINAL_DATE_TIME_FORMAT = '%Y-%m-%d'

    def __add__(self, other) -> 'Series':
        _self = self.astype('timestamp')
        _other = other.copy()

        result = _self + _other
        # current result is timestamp dtype, cast it back to date
        # (without explicitly casting db type, as time units are required for rounding)
        result = result.copy_override_type(SeriesDate)
        return self._cast_and_round_to_date(result)

    def __sub__(self, other) -> 'Series':
        # python will raise a TypeError when trying arithmetic operations
        # that are not date or timedelta type

        if other.dtype not in ('date', 'timedelta'):
            raise TypeError(f'arithmetic operation sub not supported for '
                            f'{self.__class__} and {other.__class__}')

        _self = self.astype('timestamp')
        _other = other.astype('timestamp') if other.dtype == 'date' else other.copy()

        result = _self - _other
        if other.dtype == 'date':
            # will return an interval
            return result

        # current result is timestamp dtype, cast it back to date
        # (without explicitly casting db type, as time units are required for rounding)
        result = result.copy_override_type(SeriesDate)
        return self._cast_and_round_to_date(result)


class SeriesTime(SeriesAbstractDateTime):
    """
    A Series that represents the date time and its specific operations


    **Database support and types**

    * Postgres: utilizes the 'time without time zone' database type.
    * BigQuery: utilizes the 'TIME' database type.
    * Athena: utilizes the 'double' database type. Storing the number of seconds since `00:00:00`.

    """
    dtype = 'time'
    dtype_aliases: Tuple[DtypeOrAlias, ...] = tuple()
    supported_db_dtype = {
        DBDialect.POSTGRES: 'time without time zone',
        DBDialect.BIGQUERY: 'TIME',
        # Athena supports time database type, however we decided to represent
        # time values as floats (total seconds) due to the following reasons:
        # 1. Athena does not support writing time values into tables
        # 2. Time values have till millisecond precision, using floats we can support till microseconds.
        DBDialect.ATHENA: None,
    }
    supported_value_types = (datetime.datetime, numpy.datetime64, datetime.time, str)
    supported_source_dtypes = ('string', 'timestamp, ')

    _VALID_DATE_TIME_FORMATS = SeriesAbstractDateTime._VALID_DATE_TIME_FORMATS + ['%H:%M:%S', '%H:%M:%S.%f']
    _FINAL_DATE_TIME_FORMAT = '%H:%M:%S.%f'

    @classmethod
    def get_db_dtype(cls, dialect: Dialect) -> Optional[str]:
        if is_athena(dialect):
            from bach.series import SeriesString
            return SeriesFloat64.get_db_dtype(dialect)
        return super().get_db_dtype(dialect)

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        if is_postgres(dialect) or is_bigquery(dialect):
            return super().supported_literal_to_expression(dialect=dialect, literal=literal)
        if is_athena(dialect):
            return SeriesFloat64.supported_literal_to_expression(dialect=dialect, literal=literal)
        raise DatabaseNotSupportedException(dialect)

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: AllSupportedDateTimeTypes,
        dtype: StructuredDtype
    ) -> Expression:
        val_is_null = value is None or (isinstance(value, numpy.datetime64) and numpy.isnat(value))
        if not is_athena(dialect) or val_is_null:
            return super().supported_value_to_literal(dialect=dialect, value=value, dtype=dtype)

        dt_value: Union[datetime.datetime, datetime.date, datetime.time, None] = None

        if isinstance(value, str):
            # get the right date format of the string
            str_format = cls._get_date_format_for_literal_value(value)
            dt_value = datetime.datetime.strptime(value, str_format)

        if isinstance(value, numpy.datetime64):
            dt_value = _convert_numpy_datetime_to_utc_datetime(value)

        if isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
            dt_value = value

        if not dt_value or not isinstance(value, cls.supported_value_types):
            raise ValueError(f'Not a valid {cls.dtype} literal: {value}')

        total_seconds = 0.
        if isinstance(dt_value, (datetime.datetime, datetime.time)):
            total_seconds = (
                dt_value.hour * _TOTAL_SECONDS_PER_DATE_PART[DatePart.HOUR]
                + dt_value.minute * _TOTAL_SECONDS_PER_DATE_PART[DatePart.MINUTE]
                + dt_value.second
                + dt_value.microsecond * _TOTAL_SECONDS_PER_DATE_PART[DatePart.MICROSECOND]
            )

        from bach.series import SeriesFloat64
        return SeriesFloat64.supported_value_to_literal(dialect, value=total_seconds, dtype='float64')

    def to_pandas_info(self) -> Optional[ToPandasInfo]:
        if is_athena(self.engine):
            return ToPandasInfo('object', self._parse_time_in_seconds_value_athena)
        return None

    @staticmethod
    def _parse_time_in_seconds_value_athena(value) -> Optional[datetime.time]:
        if value is None:
            return value

        hours_w_fract = value / _TOTAL_SECONDS_PER_DATE_PART[DatePart.HOUR]
        normalized_hours = int(hours_w_fract // 1)

        minutes_w_fract = (hours_w_fract - normalized_hours) * 60
        normalized_minutes = int(minutes_w_fract // 1)

        # round till milliseconds
        seconds_w_fract = round((minutes_w_fract - normalized_minutes) * 60, 6)
        normalized_seconds = int(seconds_w_fract // 1)
        return datetime.time(
            hour=normalized_hours,
            minute=normalized_minutes,
            second=normalized_seconds,
            microsecond=int(seconds_w_fract % 1 / _TOTAL_SECONDS_PER_DATE_PART[DatePart.MICROSECOND]),
        )

    # python supports no arithmetic on Time


class SeriesTimedelta(SeriesAbstractDateTime):
    """
    A Series that represents the timedelta type and its specific operations

    **Database support and types**

    * Postgres: utilizes the 'interval' database type.
    * BigQuery: utilizes the 'interval' database type.
    * Athena: utilizes the 'double' database type.
    """

    dtype = 'timedelta'
    dtype_aliases = ('interval',)
    supported_db_dtype = {
        DBDialect.POSTGRES: 'interval',
        DBDialect.BIGQUERY: 'INTERVAL',
        # Athena supports interval database type, however we decided to represent
        # timedelta values as floats (total seconds) as Athena does not support writing
        # interval values into tables
        DBDialect.ATHENA: None,
    }
    supported_value_types = (datetime.timedelta, numpy.timedelta64, str)
    supported_source_dtypes = ('string',)

    @classmethod
    def get_db_dtype(cls, dialect: Dialect) -> Optional[str]:
        if is_athena(dialect):
            return SeriesFloat64.get_db_dtype(dialect)
        return super().get_db_dtype(dialect)

    def astype(self, dtype: Union[str, Type]) -> Series:
        if get_series_type_from_dtype(dtype) is not SeriesString:
            return super().astype(dtype=dtype)

        days_series = (self.dt.total_seconds // _TOTAL_SECONDS_PER_DATE_PART[DatePart.DAY]).astype('int64')

        norm_time_ts = (
            self.dt.total_seconds - days_series * _TOTAL_SECONDS_PER_DATE_PART[DatePart.DAY]
        ).copy_override_type(SeriesFloat64)

        time_fmt_str = SeriesTimestamp.from_total_seconds(norm_time_ts).dt.strftime('%H:%M:%S.%f')

        expression = Expression.construct("{} || ' days, ' || {}", days_series.astype('string'), time_fmt_str)
        return self.copy_override(expression=expression, dtype='string')

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: AllSupportedDateTimeTypes,
        dtype: StructuredDtype
    ) -> Expression:
        if value is None:
            return Expression.construct('NULL')

        if isinstance(value, cls.supported_value_types):
            # pandas.Timedelta checks already that the string has the correct format
            # round it up to microseconds precision in order to avoid problems with BigQuery
            # pandas by default uses nanoseconds precision
            value_td = pandas.Timedelta(value).round(freq='us')

            if value_td is pandas.NaT:
                return Expression.construct('NULL')

            if is_athena(dialect):
                # athena does not support interval type, therefore we need to use epoch instead
                from bach.series import SeriesFloat64
                return SeriesFloat64.supported_value_to_literal(
                    dialect, value=value_td.total_seconds(), dtype='float64'
                )

            # interval values in iso format are allowed in SQL (both BQ and PG)
            # https://www.postgresql.org/docs/8.4/datatype-datetime.html#:~:text=interval%20values%20can%20also%20be%20written%20as%20iso%208601%20time%20intervals%2C
            return Expression.string_value(value_td.isoformat())

        raise ValueError(f'Not a valid {cls.dtype} literal: {value}')

    def to_pandas_info(self) -> Optional[ToPandasInfo]:
        if is_bigquery(self.engine):
            return ToPandasInfo(dtype='object', function=self._parse_interval_bigquery)

        if is_athena(self.engine):
            return ToPandasInfo(
                dtype='object', function=lambda value: pandas.Timedelta(seconds=value).round(freq='us')
            )

        return None

    def _parse_interval_bigquery(self, value: Optional[Any]) -> Optional[pandas.Timedelta]:
        if value is None:
            return None

        # BigQuery returns a MonthDayNano object
        # we need to normalize months to days (1 month == 30 day period)
        return pandas.Timedelta(
            days=value.days + value.months * 30,
            nanoseconds=value.nanoseconds,
        )

    def _comparator_operation(self, other, comparator,
                              other_dtypes=('timedelta', 'string'),
                              strict_other_dtypes=tuple()) -> SeriesBoolean:
        return super()._comparator_operation(other, comparator, other_dtypes, strict_other_dtypes)

    def __add__(self, other) -> 'Series':
        if isinstance(other, datetime.timedelta) or other.dtype == 'timedelta':
            return self._arithmetic_operation(
                other,
                'add',
                '({}) + ({})',
                other_dtypes=('timedelta', ),
                dtype={'timedelta': 'timedelta'}
            )
        # inverse the operands, so we call SeriesDate or SeriesTimestamp __add__
        return other + self

    def __sub__(self, other) -> 'Series':
        type_mapping = {
            'timedelta': 'timedelta',
        }
        return self._arithmetic_operation(other, 'sub', '({}) - ({})',
                                          other_dtypes=tuple(type_mapping.keys()),
                                          dtype=type_mapping)

    def __mul__(self, other) -> 'Series':
        return self._arithmetic_operation(other, 'mul', '({}) * ({})', other_dtypes=('int64', 'float64'))

    def __truediv__(self, other) -> 'Series':
        return self._arithmetic_operation(other, 'div', '({}) / ({})', other_dtypes=('int64', 'float64'))

    @property
    def dt(self) -> TimedeltaOperation:
        """
        Get access to date operations.

        .. autoclass:: bach.series.series_datetime.TimedeltaOperation
            :members:

        """
        return TimedeltaOperation(self)

    def sum(self, partition: WrappedPartition = None,
            skipna: bool = True, min_count: int = None) -> 'SeriesTimedelta':
        """
        :meta private:
        """
        result = self._derived_agg_func(
            partition=partition,
            expression='sum',
            skipna=skipna,
            min_count=min_count
        )
        return result.copy_override_type(SeriesTimedelta)

    def mean(self, partition: WrappedPartition = None, skipna: bool = True) -> 'SeriesTimedelta':
        """
        :meta private:
        """
        result = self._derived_agg_func(
            partition=partition,
            expression='avg',
            skipna=skipna
        )
        result = result.copy_override_type(SeriesTimedelta)

        if is_bigquery(self.engine):
            result = result._remove_nano_precision_bigquery()

        return result

    def _remove_nano_precision_bigquery(self) -> 'SeriesTimedelta':
        """
        Helper function that removes nano-precision from intervals.
        """
        series = self.copy()
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#interval_type
        _BQ_INTERVAL_FORMAT = '%d-%d %d %d:%d:%d.%06.0f'
        _BQ_SUPPORTED_INTERVAL_PARTS = [
            'YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND'
        ]

        # aggregating intervals by average might generate a result with
        # nano-precision, which is not supported by BigQuery TimeStamps
        # therefore we need to make sure we always generate values up to
        # microseconds precision
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#timestamp_type
        all_extracted_parts_expr = [
            Expression.construct(f'EXTRACT({date_part} FROM {{}})', series)
            for date_part in _BQ_SUPPORTED_INTERVAL_PARTS
        ]
        # convert nanoseconds to microseconds
        all_extracted_parts_expr.append(
            Expression.construct(f'EXTRACT(NANOSECOND FROM {{}}) / 1000', series)
        )
        format_arguments_expr = join_expressions(all_extracted_parts_expr)

        # All parts will create a string with following format
        # '%d-%d %d %d:%d:%d.%06.0f'
        # where the first 6 digits are date parts from YEAR to SECOND
        # Format specifier %06.0f will format fractional part of seconds with maximum width of 6 digits
        # for example:
        # nanoseconds = 1142857, converting them into microseconds is 1142.857
        # when applying string formatting, the value will be rounded into 1143 (.0 precision)
        # and will be left padded by 2 leading zeros: 001143 (0 flag and 6 minimum width)
        # for more information:
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/string_functions#format_string
        format_expr = Expression.construct(
            f'format({{}}, {{}})',
            Expression.string_value(_BQ_INTERVAL_FORMAT),
            format_arguments_expr,
        )
        return series.copy_override(
            expression=self.dtype_to_expression(
                self.engine, source_dtype='string', expression=format_expr,
            )
        )

    def quantile(
        self, partition: WrappedPartition = None, q: Union[float, List[float]] = 0.5,
    ) -> 'SeriesTimedelta':
        """
        When q is a float or len(q) == 1, the resultant series index will remain
        In case multiple quantiles are calculated, the resultant series index will have all calculated
        quantiles as index values.
        """
        result = self.to_frame().quantile(q=q, partition=partition)[f'{self.name}_quantile']
        result = result.copy_override(name=self.name)

        return result

    def mode(self, partition: WrappedPartition = None, skipna: bool = True) -> 'SeriesTimedelta':
        if not is_bigquery(self.engine):
            return super().mode(partition, skipna)

        # APPROX_TOP_COUNT does not support INTERVALS. So, we should calculate the mode based
        # on the total seconds
        total_seconds_mode = self.dt.total_seconds.mode(partition, skipna)
        return self.from_total_seconds(total_seconds_mode)

    @classmethod
    def from_total_seconds(cls, total_seconds: SeriesAbstractNumeric) -> 'SeriesTimedelta':
        """
        Converts a numerical series representing total seconds (epoch) to timedelta series.

        returns a SeriesTimedelta
        """
        engine = total_seconds.engine
        if is_athena(engine):
            # no action is required for athena, as we represent timedelta with total seconds by default.
            return total_seconds.copy_override_type(SeriesTimedelta)

        seconds_as_timestamp_series = SeriesTimestamp.from_total_seconds(total_seconds)

        start_time = SeriesTimestamp.from_value(
            base=seconds_as_timestamp_series, value=datetime.datetime(1970, 1, 1)
        )

        return (seconds_as_timestamp_series - start_time).copy_override_type(SeriesTimedelta)


def _convert_numpy_datetime_to_utc_datetime(value: numpy.datetime64) -> datetime.datetime:
    # Weird trick: count number of microseconds in datetime, but only works on timedelta, so convert
    # to a timedelta first, by subtracting 0 (epoch = 1970-01-01 00:00:00)
    # Rounding can be unpredictable because of limited precision, so always truncate excess precision
    microseconds = int((value - numpy.datetime64('1970', 'us')) // numpy.timedelta64(1, 'us'))
    return datetime.datetime.utcfromtimestamp(microseconds / 1_000_000)



"""
Copyright 2021 Objectiv B.V.
"""
import typing
from typing import Dict, TypeVar, Tuple, List, Optional, Mapping, Hashable, Union

from sqlalchemy.engine import Dialect

from bach.expression import Expression, get_variable_tokens, VariableToken
from bach.types import value_to_dtype, get_series_type_from_dtype
from sql_models.util import quote_identifier
from sql_models.model import CustomSqlModelBuilder, SqlModel, Materialization, SqlModelSpec
from sql_models.constants import NotSet, not_set

T = TypeVar('T', bound='SqlModelSpec')
TBachSqlModel = TypeVar('TBachSqlModel', bound='BachSqlModel')


if typing.TYPE_CHECKING:
    from bach.dataframe import DtypeNamePair


class BachSqlModel(SqlModel[T]):
    """
    SqlModel with meta information about the columns that it produces.
    This additional information needs to be specifically set at model instantiation, it cannot be deduced
    from the sql.

    The column information is not used for sql generation, but can be used by other code
    interacting with the models. The information is not reflected in the `hash`, as it doesn't matter for
    the purpose of sql generation.
    """
    def __init__(
        self,
        model_spec: T,
        placeholders: Mapping[str, Hashable],
        references: Mapping[str, 'SqlModel'],
        materialization: Materialization,
        materialization_name: Optional[str],
        column_expressions: Dict[str, Expression],
    ) -> None:
        """
        Similar to :py:meth:`SqlModel.__init__()`. With one additional parameter: column_expressions,
        a mapping between the names of the Bach Series and expressions that this model's query will return in
        the correct order.

        Note that they keys in column_expressions are the names as used in Bach, not the names as they appear
        in the sql query. If a column name contains special characters, then the sql might use an escaped
        name. See :meth:`columns` for more information.
        """
        self._column_expressions = column_expressions
        super().__init__(
            model_spec=model_spec,
            placeholders=placeholders,
            references=references,
            materialization=materialization,
            materialization_name=materialization_name,
        )

    @property
    def series_names(self) -> Tuple[str, ...]:
        # TODO: rename this, series_names? columns is confusing
        """
        Names of the Series that the query of this model returns, in order.

        Note: These are the names as they would appear in a Bach DataFrame. Columns might have a different
        name in the sql query of this model than they have in a Bach DataFrame, depending on what characters
        the database supports in column names. Example: `df['X!@'] = 1` might lead to sql of the form:
        'select 1 as escaped_column_name from ...' instead of 'select 1 as X!@'. This function will return
        the name used in the Bach DataFrame, so in the case of the example 'X!@'.
        """
        return tuple(self._column_expressions.keys())

    @property
    def column_expressions(self) -> Dict[str, Expression]:
        """
        Mapping containing the expression used per Bach series-name.

        Note: The series names are as they would appear in a bach DataFrame. Within the sql query of this
        model the columns might be called differently. See the description of :meth:`columns` for more
        information.
        """
        return self._column_expressions

    def copy_override(
        self: TBachSqlModel,
        *,
        model_spec: T = None,
        placeholders: Mapping[str, Hashable] = None,
        references: Mapping[str, 'SqlModel'] = None,
        materialization: Materialization = None,
        materialization_name: Union[Optional[str], NotSet] = not_set,
        column_expressions: Dict[str, Expression] = None
    ) -> TBachSqlModel:
        """
        Similar to super class's implementation, but adds optional 'columns' parameter
        """
        materialization_name_value = (
            self.materialization_name if materialization_name is not_set else materialization_name
        )
        return self.__class__(
            model_spec=self.model_spec if model_spec is None else model_spec,
            placeholders=self.placeholders if placeholders is None else placeholders,
            references=self.references if references is None else references,
            materialization=self.materialization if materialization is None else materialization,
            materialization_name=materialization_name_value,
            column_expressions=self.column_expressions if column_expressions is None else column_expressions
        )

    @classmethod
    def from_sql_model(cls, sql_model: SqlModel, column_expressions: Dict[str, Expression]) -> 'BachSqlModel':
        """ From any SqlModel create a BachSqlModel with the given column definitions. """
        return cls(
            model_spec=sql_model.model_spec,
            placeholders=sql_model.placeholders,
            references=sql_model.references,
            materialization=sql_model.materialization,
            materialization_name=sql_model.materialization_name,
            column_expressions=column_expressions,
        )

    @classmethod
    def _get_placeholders(
        cls,
        dialect: Dialect,
        variables: Dict['DtypeNamePair', Hashable],
        expressions: List[Expression],
    ) -> Dict[str, str]:
        filtered_variables = filter_variables(variables, expressions)
        return get_variable_values_sql(dialect, filtered_variables)


class SampleSqlModel(BachSqlModel):
    """
    A custom SqlModel that simply adds an extra property to the BachSqlModel: previous.

    The previous property is not used in the generated sql at all, but can be used to track a previous
    SqlModel. This is useful for how we implemented sampling, as that effectively inserts a sql-model in the
    graph that has no regular reference to the previous node in the graph. By storing the previous node
    here, we can later still reconstruct what the actual previous node was with some custom logic.

    See the DataFrame.sample() implementation for more information
    """
    def __init__(
        self,
        model_spec: T,
        placeholders: Mapping[str, Hashable],
        references: Mapping[str, 'SqlModel'],
        materialization: Materialization,
        materialization_name: Optional[str],
        column_expressions: Dict[str, Expression],
        previous: BachSqlModel,
    ) -> None:
        self.previous = previous
        super().__init__(
            model_spec=model_spec,
            placeholders=placeholders,
            references=references,
            materialization=materialization,
            materialization_name=materialization_name,
            column_expressions=column_expressions,
        )

    def copy_override(
        self: 'SampleSqlModel',
        *,
        model_spec: T = None,
        placeholders: Mapping[str, Hashable] = None,
        references: Mapping[str, 'SqlModel'] = None,
        materialization: Materialization = None,
        materialization_name: Union[Optional[str], NotSet] = not_set,
        column_expressions: Dict[str, Expression] = None,
        previous: BachSqlModel = None
    ) -> 'SampleSqlModel':
        """
        Similar to super class's implementation, but adds optional 'previous' parameter
        """
        materialization_name_value = \
            self.materialization_name if materialization_name is not_set else materialization_name
        return self.__class__(
            model_spec=self.model_spec if model_spec is None else model_spec,
            placeholders=self.placeholders if placeholders is None else placeholders,
            references=self.references if references is None else references,
            materialization=self.materialization if materialization is None else materialization,
            materialization_name=materialization_name_value,
            column_expressions=self.column_expressions if column_expressions is None else column_expressions,
            previous=self.previous if previous is None else previous
        )

    @classmethod
    def from_bach_sql_model(cls, model: BachSqlModel, previous: BachSqlModel) -> 'BachSqlModel':
        """ From any BachSqlModel create a SampleSqlModel with the given `previous` model. """
        return cls(
            model_spec=model.model_spec,
            placeholders=model.placeholders,
            references=model.references,
            materialization=model.materialization,
            materialization_name=model.materialization_name,
            column_expressions=model.column_expressions,
            previous=previous
        )


class CurrentNodeSqlModel(BachSqlModel):
    @staticmethod
    def get_instance(
        *,
        dialect: Dialect,
        name: str,
        column_names: Tuple[str, ...],
        column_exprs: List[Expression],
        distinct: bool,
        where_clause: Optional[Expression],
        group_by_clause: Optional[Expression],
        having_clause: Optional[Expression],
        order_by_clause: Optional[Expression],
        limit_clause: Expression,
        previous_node: BachSqlModel,
        variables: Dict['DtypeNamePair', Hashable],
    ) -> 'CurrentNodeSqlModel':

        columns_str = ', '.join(expr.to_sql(dialect) for expr in column_exprs)
        distinct_stmt = ' distinct ' if distinct else ''
        where_str = where_clause.to_sql(dialect) if where_clause else ''
        group_by_str = group_by_clause.to_sql(dialect) if group_by_clause else ''
        having_str = having_clause.to_sql(dialect) if having_clause else ''
        order_by_str = order_by_clause.to_sql(dialect) if order_by_clause else ''
        limit_str = limit_clause.to_sql(dialect) if limit_clause else ''

        sql = (
            f"select {distinct_stmt}{columns_str} \n"
            f"from {{{{prev}}}} \n"
            f"{where_str} \n"
            f"{group_by_str} \n"
            f"{having_str} \n"
            f"{order_by_str} \n"
            f"{limit_str} \n"
        )

        # Add all references found in the Expressions to self.references
        nullable_expressions = [where_clause, group_by_clause, having_clause, order_by_clause, limit_clause]
        all_expressions = column_exprs + [expr for expr in nullable_expressions if expr is not None]
        references = construct_references({'prev': previous_node}, all_expressions)

        return CurrentNodeSqlModel(
            model_spec=CustomSqlModelBuilder(sql=sql, name=name),
            placeholders=BachSqlModel._get_placeholders(dialect, variables, all_expressions),
            references=references,
            materialization=Materialization.CTE,
            materialization_name=None,
            column_expressions={name: expr for name, expr in zip(column_names, column_exprs)},
        )


def construct_references(
        base_references: Mapping[str, 'SqlModel'],
        expressions: List['Expression']
) -> Dict[str, 'SqlModel']:
    """
    Create a dictionary of references consisting of the base_references and all references found in the
    expressions.

    Will raise an exception if there are references with the same name that reference different models.
    """
    result: Dict[str, SqlModel] = {}
    for expr in expressions:
        references = expr.get_references()
        _check_reference_conflicts(result, references)
        result.update(references)
    _check_reference_conflicts(base_references, result)
    result.update(base_references)
    return result


def _check_reference_conflicts(left: Mapping[str, 'SqlModel'], right: Mapping[str, 'SqlModel']) -> None:
    """
    Util function: Check that two dicts with references don't have conflicting values.
    """
    for ref_name, model in right.items():
        if left.get(ref_name) not in (None, model):
            # This should never happen, if other code doesn't mess up.
            # We have this check as a backstop assertion to fail early
            raise Exception(f'Encountered reference {ref_name} before, but with a different value: '
                            f'{left.get(ref_name)} != {model}')


def filter_variables(
        variable_values: Dict['DtypeNamePair', Hashable],
        filter_expressions: List['Expression']
) -> Dict['DtypeNamePair', Hashable]:
    """
    Util function: Return a copy of the variable_values, with only the variables for which there is a
    VariableToken in the filter_expressions.
    """
    available_tokens = get_variable_tokens(filter_expressions)
    dtype_names = {token.dtype_name for token in available_tokens}

    return {dtype_name: value for dtype_name, value in variable_values.items() if dtype_name in dtype_names}


def get_variable_values_sql(
        dialect: Dialect,
        variable_values: Dict['DtypeNamePair', Hashable]
) -> Dict[str, str]:
    """
    Take a dictionary with variable_values and return a dict with the full variable names and the values
    as sql.
    The sql assumes it will be used as values for SqlModels's placeholders. i.e. It will not be format
    escaped, unlike if it would be used directly into SqlModel.sql in which case it would be escaped twice.
    The sql will be proper sql tho, with identifier, strings, etc. properly quoted and escaped.

    :param variable_values: Mapping of variable to value.
    :return: Dictionary mapping full variable name to sql literal
    """
    result = {}
    for dtype_name, value in variable_values.items():
        dtype, name = dtype_name
        value_dtype = value_to_dtype(value)
        if dtype != value_dtype:  # should never happen
            Exception(f'Dtype of value {value}, {value_dtype} does not match registered dtype {dtype}')
        placeholder_name = VariableToken.dtype_name_to_placeholder_name(dtype=dtype, name=name)
        series_type = get_series_type_from_dtype(dtype)
        expr = series_type.value_to_literal(dialect=dialect, value=value, dtype=dtype)
        double_escaped_sql = expr.to_sql(dialect)
        sql = double_escaped_sql.format().format()
        result[placeholder_name] = sql
    return result



from enum import Enum
from typing import Union

from sqlalchemy.engine import Dialect, Engine


class NotSet(Enum):
    """
    INTERNAL: Special token used as default value for parameters, to distinguish the default value from
    None for Optional values.
    """
    token = 0


not_set: NotSet = NotSet.token


class DBDialect(Enum):
    """
    Supported database dialects.

    The values are equal to Dialect.name for specific SqlAlchemy Dialect classes e.g.
        DBDialects.POSTGRES.value == PGDialect.name

    Generally we'll use the sqlalchemy.engine.Dialect class to define a certain database type (e.g. when
    passing as a parameter). However, when needing to specify a database type in generic code, then use this
    class. This class hardcodes the values, and thus does not depend on any specific python packages being
    installed. i.e. accessing DBDialect.BIGQUERY works even if the 'sqlalchemy-bigquery' package is not
    installed.
    """
    POSTGRES = 'postgresql'  # value of PGDialect.name
    BIGQUERY = 'bigquery'  # value of BigQueryDialect.name
    ATHENA = 'awsathena'  # value of AthenaDialect.name

    def is_dialect(self, dialect_engine: Union[Dialect, Engine]) -> bool:
        return dialect_engine.name == self.value

    @classmethod
    def from_engine(cls, engine: Engine) -> 'DBDialect':
        return DBDialect(engine.name)

    @classmethod
    def from_dialect(cls, dialect: Dialect) -> 'DBDialect':
        return DBDialect(dialect.name)



"""
Copyright 2021 Objectiv B.V.
"""
from sql_models.graph_operations import get_graph_nodes_info, NodeInfo
from sql_models.model import SqlModel


def print_graph_info(model: SqlModel):
    for node_info in get_graph_nodes_info(model):
        print_node_info(node_info)


def print_node_info(node_info: NodeInfo):
    print('\n----------')
    print(f'name: {node_info.node_id}')
    print(f'hash: {node_info.model.hash}')
    print(f'reference_path: {node_info.reference_path}')
    print(f'placeholder values: {node_info.model.placeholders}')
    print(f'materialization: {node_info.model.materialization}')
    print(f'materialization_name: {node_info.model.materialization_name}')
    print(f'inputs: {[in_node.node_id for in_node in node_info.in_edges]}')
    print(f'outputs: {[out_node.node_id for out_node in node_info.out_edges]}')



"""
Copyright 2021 Objectiv B.V.
"""
import re
from typing import Union, TYPE_CHECKING, Optional, Pattern

from sqlalchemy.engine import Dialect

from bach.partitioning import get_order_by_expression
from bach.series import Series
from bach.expression import Expression, AggregateFunctionExpression, get_variable_tokens, ConstValueExpression
from bach.series.series import WrappedPartition
from bach.types import StructuredDtype, get_series_type_from_dtype
from sql_models.constants import DBDialect
from sql_models.util import DatabaseNotSupportedException, is_bigquery, is_postgres, is_athena

if TYPE_CHECKING:
    from bach.series import SeriesBoolean, SeriesInt64
    from bach import DataFrame, SortColumn, SeriesJson


class StringOperation:

    def __init__(self, base: 'SeriesString'):
        self._base = base

    def __getitem__(self, key: Union[int, slice]) -> 'SeriesString':
        """
        Get a python string slice using DB functions. Format follows standard slice format
        :param key: an int for a single character, or a slice for some nice slicing
        :return: SeriesString with the slice applied
        """
        if key and not isinstance(key, (int, slice)):
            raise TypeError(f'Type not supported {type(key)}')

        start = key.start if isinstance(key, slice) else key

        if isinstance(key, slice):
            stop = key.stop
        else:
            # if key == -1, then we should just return last character, otherwise stop on the next character
            stop = key + 1 if key != -1 else None

        return self._base.copy_override(expression=self._get_substr_expression(start=start, stop=stop))

    def _get_substr_expression(self, start=None, stop=None) -> Expression:
        base_expr = self._base.expression
        len_series = self._base.str.len()

        # value[:]
        if start is None and stop is None:
            return base_expr

        # For value[x:y], where  x and y have the same sign and x > y > 0.
        # result will always be empty str
        if start and stop and start * stop > 0 and start >= stop:
            return Expression.string_value('')

        # get the equivalent SQL index for each PYTHON slice index
        start_index_expr = self._python_index_to_sql_index_expr(start)
        stop_index_expr = self._python_index_to_sql_index_expr(stop)

        if stop is None:
            substr_len_expr = len_series.expression
        else:
            substr_len_expr = Expression.construct(
                'greatest(({}) - ({}), 0)',
                stop_index_expr, start_index_expr,
            )

        return Expression.construct('substr({}, {}, {})', base_expr, start_index_expr, substr_len_expr)

    def _python_index_to_sql_index_expr(self, index: Optional[int]) -> Expression:
        """
        returns expression for converting pythonic index to SQL ordinal index.

        .. note::
            if provided index is None, then SQL index will default to 1.

        .. note::
           if provided index exceeds negative range (|index| > len(value)), then SQL index
           will default to 1.
        """
        # converts python index to its equivalent SQL (positive) index.
        # for example:
        #
        #          -8  -7  -6  -5  -4  -3  -2  -1
        #           O | B | J | E | C | T | I | V
        #  sql      1   2   3   4   5   6   7   8
        # python    0   1   2   3   4   5   6   7

        #
        # python     sql
        #   0         1
        #  -6         3
        # -10         1

        if index is None or index >= 0:
            sql_index = 0 if index is None else index
            index_expr = ConstValueExpression.construct(str(sql_index))
        # if index is negative, we need to get its respective positive index
        else:
            index_expr = Expression.construct(f'greatest({{}} - {abs(index)}, 0)', self.len())

        # SQL starts from 1, therefore we need to normalize python start index
        return Expression.construct('{} + 1', index_expr)

    def slice(self, start=None, stop=None) -> 'SeriesString':
        """
        slice a string like you would in Python, either by calling this method, or by slicing directly
        on the `str` accessor.

        .. code-block:: python

            a.str[3]            # get one char
            a.str[3:5]          # get a slice from char 3-5
            a.str.slice(3, 5)   # idem
        """
        if isinstance(start, slice):
            return self.__getitem__(start)
        return self.__getitem__(slice(start, stop))

    def replace(
        self,
        pat: Union[str, Pattern],
        repl: str,
        n: int = -1,
        case: bool = None,
        flags: int = 0,
        regex: bool = False,
    ) -> 'SeriesString':
        """
        replace each occurrence of a pattern in SeriesString.

        :param pat: string or compiled pattern to use for replacement.
        :param repl: string to use as a replacement for each occurrence
        :param n: number of occurrences to be replaced. Only n=-1 is supported (all occurrences)
        :param case: determines if the replace is case insensitive. Considered only when regex=True.
        :param flags: regex module flags. Considered only when regex=True.
        :param regex: Determines if provided pattern is a regular expression or not.

        .. note::
          Replacements based on regular expressions are not supported yet. Therefore:
              - `pat` parameter must be string type
              - `case`, `flags`, `regex` will not be considered on replacement
        """
        if isinstance(pat, re.Pattern) or regex:
            raise NotImplementedError('Regex patterns are not supported yet.')

        if n != -1:
            raise NotImplementedError('Replacement for all occurrences is supported only.')

        expr = Expression.construct(
            f'REPLACE({{}}, {{}}, {{}})',
            self._base,
            Expression.string_value(pat),
            Expression.string_value(repl),
        )
        return self._base.copy_override(expression=expr)

    def upper(self) -> 'SeriesString':
        """
        converts string values into uppercase.

        :return: SeriesString with all alphabetic characters in uppercase
        """
        return self._base.copy_override(
            expression=Expression.construct('upper({})', self._base)
        )

    def lower(self) -> 'SeriesString':
        """
        converts string values into lowercase.

        :return: SeriesString with all alphabetic characters in lowercase
        """
        return self._base.copy_override(
            expression=Expression.construct('lower({})', self._base)
        )

    def len(self) -> 'SeriesInt64':
        """
        gets the lengths of string values.

        :return: SeriesInt64
        """
        from bach.series import SeriesInt64
        return self._base.copy_override(
            expression=Expression.construct('length({})', self._base)
        ).copy_override_type(SeriesInt64)


class SeriesString(Series):
    """
    A Series that represents the string type and its specific operations

    **Operations**

    Strings can be concatenated using the '+' operator, and the 'str' accessor can be used to get access
    to slices.

    Example:

    .. code-block:: python

        c = a + b  # concat the strings.
        a.str[3]   # get one char
        a.str[3:5] # get a slice from char 3-5


    **Database support and types**

    * Postgres: utilizes the 'text' database type.
    * Athena: utilizes the 'varchar' database type.
    * BigQuery: utilizes the 'STRING' database type.
    """

    dtype = 'string'
    dtype_aliases = ('text', str)
    supported_db_dtype = {
        DBDialect.POSTGRES: 'text',
        DBDialect.ATHENA: 'varchar',
        DBDialect.BIGQUERY: 'STRING'
    }
    supported_value_types = (str, type(None))  # NoneType ends up as a string for now

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        # We override the parent class here because strings are really common, and we don't strictly need
        # to cast them. As all supported databases will interpret a string literal as a string.
        # Not casting string literals greatly improves the readability of the generated SQL.

        # However, there is an edge case: NULL values should be cast to string. e.g. BigQuery considers a
        # naked NULL to be INT64. Additionally, we'll always cast variables, just so this keeps working if
        # the variable get set to `None`
        if literal.to_sql(dialect=dialect).upper() == 'NULL' or get_variable_tokens([literal]):
            return super().supported_literal_to_expression(dialect=dialect, literal=literal)
        return literal

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: str,
        dtype: StructuredDtype
    ) -> Expression:
        return Expression.string_value(value)

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        if source_dtype == 'string':
            return expression

        cast_expr = Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)
        series_type = get_series_type_from_dtype(source_dtype)
        from bach.series import SeriesJson, SeriesFloat64

        if is_athena(dialect):
            if issubclass(series_type, SeriesJson):
                # casting directly to varchar will "deserialize" the json text, meaning that the string value
                # will be extracted as scalar. For example: "a string" will result into: a string
                # which yields different results compare to Postgres and BigQuery
                # Using json_format will avoid this and double quotes will be kept
                return Expression.construct('json_format({})', expression)

            if issubclass(series_type, SeriesFloat64):
                # casting doubles to varchar do not trim trailing zeros after decimal point
                # (if value has no fractional part)
                # for example, 123.0 will result into '123.0' (this does not happen for PG and BQ),
                # so for data consistency, we should trim the extra .0
                return Expression.construct(r"regexp_replace({},'\.0+$','')", cast_expr)

        return cast_expr

    def get_dummies(
        self,
        prefix: Optional[str] = None,
        prefix_sep: str = '_',
        dummy_na: bool = False,
        dtype: str = 'int64',
    ) -> 'DataFrame':
        """
        Convert each unique category/value from the series into a dummy/indicator variable.

        :param prefix: String to append to each new column name. By default, the prefix will be the name of
            the caller.
        :param prefix_sep: Separated between the prefix and label.
        :param dummy_na: If true, it will include ``nan`` as a variable.
        :param dtype: dtype of all new columns

        :return: DataFrame

        .. note::
            Series should contain at least one index level.
        """
        return self.to_frame().get_dummies(
            prefix=prefix, prefix_sep=prefix_sep, dummy_na=dummy_na, dtype=dtype,
        )

    @property
    def str(self) -> StringOperation:
        """
        Get access to string operations.

        .. autoclass:: bach.series.series_string.StringOperation
            :members:

        """
        return StringOperation(self)

    def __add__(self, other) -> 'Series':
        return self._binary_operation(other, 'concat', '{} || {}', other_dtypes=('string',))

    def _comparator_operation(self, other, comparator, other_dtypes=tuple(['string']),
                              strict_other_dtypes=tuple()) -> 'SeriesBoolean':
        return super()._comparator_operation(other, comparator, other_dtypes, strict_other_dtypes)

    def to_json_array(self, partition: Optional[WrappedPartition] = None) -> 'SeriesJson':
        """
        Aggregate function: Group the values of this Series into a json array

        The order of the values in the array will be based of the order of the values in this Series. If
        this Series does not have a deterministic sorting, then the values are additionally sorted by the
        values themselves. Null values will always be sorted last when aggregating all values, following
        default sorting behavior for DataFrame/Series.

        :param partition: The partition to apply, optional.
        :return: SeriesJson containing an array of strings on each row.
        """
        order_by = self.order_by
        # Add this series as the final column to sort on. If the order_by is deterministic then this won't
        # change the sorting. If order_by was not deterministic, then this will make it deterministic.
        from bach import SortColumn
        order_by += [SortColumn(expression=self.expression, asc=True)]

        order_by_expr = get_order_by_expression(
            dialect=self.engine.dialect, order_by=order_by, na_position='last',
        )
        array_agg_expression = AggregateFunctionExpression.construct('array_agg({} {})', self, order_by_expr)
        if is_postgres(self.engine):
            expression = Expression.construct('to_jsonb({})', array_agg_expression)
        elif is_bigquery(self.engine):
            expression = Expression.construct('to_json_string({})', array_agg_expression)
        elif is_athena(self.engine):
            expression = Expression.construct('cast({} as json)', array_agg_expression)
        else:
            raise DatabaseNotSupportedException(self.engine)

        result = self._derived_agg_func(partition, expression)
        from bach import SeriesJson
        return result.copy_override_type(SeriesJson)



"""
Copyright 2021 Objectiv B.V.

Models:

* `SqlModel`: Actual sql-model. Instances of this class form the nodes in the model graph
* `SqlModelSpec`: Specification class, that specifies basic properties of an SqlModel instance
* `SqlModelBuilder`: Helper class to instantiate the (immutable) SqlModel objects. Generally models should
  extend this class and not the SqlModel class.
* `CustomSqlModelBuilder`: Utility child of SqlModelSpec that can be used to add a node with custom sql to a
  model graph.
* `Materialization`: Specifies what kind of SQL should be generated: a query or a specific create statement

"""
import collections
import hashlib
from abc import abstractmethod, ABCMeta
from copy import deepcopy
from enum import Enum
from typing import TypeVar, Generic, Dict, Any, Set, Tuple, Type, Union, Hashable, NamedTuple, Optional, \
    Mapping, MutableMapping, cast

from sql_models.constants import not_set, NotSet
from sql_models.util import extract_format_fields


class MaterializationType(NamedTuple):
    """
    type_name: unique identifier
    is_statement:
        True indicates that a SqlModel with this materialization should be turned into a sql statement,
            either query or a create statement
        False indicates that a SqlModel with this materialization should not be turned into a standalone sql
            statement. Example: a CTE should not be used as a standalone statement.
    is_cte:
        True indicates that a SqlModel with this materialization can be used as a CTE by another SqlModel.
        False indicates that a SqlModel with this materialization cannot be used as a CTE
    has_lasting_effect: true iff the materialization is a permanent table or view
    """
    type_name: str
    is_statement: bool
    is_cte: bool
    has_lasting_effect: bool


class Materialization(Enum):
    CTE = MaterializationType('cte', is_statement=False, is_cte=True, has_lasting_effect=False)
    """ A QUERY can be used as a CTE, but is also a stand-alone query."""
    QUERY = MaterializationType('query', is_statement=True, is_cte=True, has_lasting_effect=False)
    VIEW = MaterializationType('view', is_statement=True, is_cte=False, has_lasting_effect=True)
    SOURCE = MaterializationType('source', is_statement=False, is_cte=False, has_lasting_effect=False)
    TABLE = MaterializationType('table', is_statement=True, is_cte=False, has_lasting_effect=True)
    """
    TEMP TABLE is a temporary table that is limited to the current session, or a table that is guaranteed to
    be cleaned up by the database at some later time.
    """
    TEMP_TABLE = MaterializationType('temp_table', is_statement=True, is_cte=False, has_lasting_effect=False)
    """ A VIRTUAL_NODE will not be turned into a statement, nor generate any CTEs"""
    VIRTUAL_NODE = MaterializationType('virtual', is_statement=False, is_cte=False, has_lasting_effect=False)

    @property
    def type_name(self) -> str:
        """ Materialization type unique name e.g. 'cte' or 'table' """
        return self.value.type_name

    @property
    def is_statement(self) -> bool:
        return self.value.is_statement

    @property
    def is_cte(self) -> bool:
        return self.value.is_cte

    @property
    def has_lasting_effect(self) -> bool:
        return self.value.has_lasting_effect

    @classmethod
    def get_by_type_name(cls, type_name) -> 'Materialization':
        for mat in cls:
            if mat.type_name == type_name:
                return mat
        raise KeyError(f'no such Materialization: "{type_name}"')

    @classmethod
    def normalize(cls, materialization: Union['Materialization', str]) -> 'Materialization':
        if isinstance(materialization, str):
            return Materialization.get_by_type_name(materialization)
        if isinstance(materialization, Materialization):
            return materialization
        raise ValueError(f'materialization must be of type str or Materialization, '
                         f'encountered type: {type(materialization)}')


# special reference-level format string that will be filled in at sql-generation time with a per-model
# unique string
REFERENCE_UNIQUE_FIELD = 'id'


RefPath = Tuple[str, ...]

T = TypeVar('T', bound='SqlModelSpec')
TB = TypeVar('TB', bound='SqlModelBuilder')
TSqlModel = TypeVar('TSqlModel', bound='SqlModel')


class SqlModelSpec:
    """
    Abstract immutable class that specifies the sql, placeholders, and references that a
    SqlModel instance should have.

    Child classes need to define the actual template by implementing the sql, placeholders, and reference
    functions.

    Generally it's better to actually extend of the subclass ComposableSqlModel, as that has some methods
    to make it easier to instantiate an SqlModel based on the Template.
    """

    def __init__(self):
        # Basic check on the child-classes implementation of spec_references and spec_placeholders
        overlap = self.spec_references.intersection(self.spec_placeholders)
        if overlap:
            raise Exception(f'Specified reference names and specified placeholder names of class overlap.'
                            f'This indicates an error in the implementation of the'
                            f'{self.__class__.__name__} class. Overlap: {overlap}')

    @property
    def generic_name(self):
        """ Generic name. Can be overridden by subclasses. Must be a constant. """
        return self.__class__.__name__

    @property
    @abstractmethod
    def sql(self):
        """ Must be implemented by child class. Return value should typically be a constant. """
        raise NotImplementedError()

    @property
    @abstractmethod
    def spec_references(self) -> Set[str]:
        """ Must be implemented by child class. Return value should typically be a constant. """
        # TODO: add type information
        raise NotImplementedError()

    @property
    @abstractmethod
    def spec_placeholders(self) -> Set[str]:
        """ Must be implemented by child class. Return value should typically be a constant. """
        raise NotImplementedError()

    @staticmethod
    def placeholders_to_sql(placeholders: Mapping[str, Any]) -> Dict[str, str]:
        """
        Child classes can override this function if some of the placeholders require conversion before being
        used in format(). Should be a constant, pure, and immutable function.

        If any format string exists, and it should not be substituted in the reference resolving step,
        it should be escaped with escape_placeholder_value.
        """
        # Override for non-default behaviour
        # If we switch to jinja templates, then we won't need this function anymore.
        return {key: escape_placeholder_value(str(val)) for key, val in placeholders.items()}

    def assert_adheres_to_spec(self,
                               references: Mapping[str, 'SqlModel'],
                               placeholders: Mapping[str, Any]):
        """
        Verify that the references and placeholder values adhere to the specifications of self.
        :raise Exception: If a reference or placeholder value is missing
        """
        spec = self
        reference_keys = set(references.keys())
        placeholder_keys = set(placeholders.keys())
        # Allow for more references to be present, as they could've been added dynamically.
        if len(spec.spec_references - reference_keys) > 0:
            raise Exception(f'Provided references for model {spec.__class__.__name__} '
                            f'do not match required references: '
                            f'{sorted(reference_keys)} != {sorted(spec.spec_references)}')
        for reference_key, reference_value in references.items():
            if not isinstance(reference_value, SqlModel):
                raise Exception(f'Provided reference for model {spec.__class__.__name__} is not an '
                                f'instance of SqlModel. '
                                f'Reference: {reference_key}, type: {type(reference_value)}')
        if placeholder_keys != spec.spec_placeholders:
            raise Exception(f'Provided placeholders for model {spec.__class__.__name__} '
                            f'do not match required placeholders: '
                            f'{sorted(placeholder_keys)} != {sorted(spec.spec_placeholders)}')


class SqlModelBuilder(SqlModelSpec, metaclass=ABCMeta):
    """
    Extension of SqlModelSpec that adds functions to easily build an instance of
    SqlModel that uses the SqlModelSpec

    There are multiple ways to create an instance of a SqlModel:
        1) use build() on this class
        2.1) use instantiate_recursively() on an instance of this class
        2.2) use instantiate() on an instance of this class
        2.3) user __call__() on an instance of this class

    A single SqlModelBuilder can be used to instantiate multiple SqlModel models. If this
    class is used multiple times to create an instance with the same placeholders, references,
    and materialization, then the same instance is returned each time.
    """

    def __init__(self, **values: Any):
        # initialize values first, so any indirect calls to references and placeholders
        # in super.__init__() will work
        self._references: Dict[str, Union[SqlModelBuilder, SqlModel]] = {}
        self._placeholders: Dict[str, Any] = {}
        super().__init__()
        self.set_values(**values)
        self.materialization = Materialization.CTE
        self.materialization_name: Optional[str] = None
        self._cache_created_instances: Dict[str, 'SqlModel'] = {}

    @property
    def spec_references(self) -> Set[str]:
        """
        Automatically determine the reference names from the sql, which contains '{{reference}}' strings.
        """
        # get all format strings that are formatted as {{x}}, but explicitly remove the
        # REFERENCE_UNIQUE_FIELD, as it is a magic value that will be filled in later.
        return extract_format_fields(self.sql, 2) - {REFERENCE_UNIQUE_FIELD}

    @property
    def spec_placeholders(self) -> Set[str]:
        """
        Automatically determine the placeholder names from the sql, which contains '{placeholder}' strings.
        """
        return extract_format_fields(self.sql)

    @property
    def references(self):
        # return shallow-copy of the dictionary.
        # keys are strings and thus immutable, values are included uncopied.
        return {key: value for key, value in self._references.items()}

    @property
    def placeholders(self):
        # return deepcopy of the dictionary
        return deepcopy(self._placeholders)

    @classmethod
    def build(cls: Type[TB], **values) -> 'SqlModel[TB]':
        """
        Class method that instantiates this SqlModelBuilder class, and uses it to
        recursively instantiate SqlModel[T].

        This might mutate referenced SqlModelBuilder objects, see instantiate_recursively()
        for more information.
        """
        builder_instance: TB = cls(**values)
        return builder_instance.instantiate_recursively()

    def instantiate_recursively(self: TB) -> 'SqlModel[TB]':
        """
        Creates an instance of SqlModel[T] like instantiate(), but unlike instantiate()
        this will convert references that are SqlModelBuilder to SqlModel too.
        And this is done in a recursive manner. Note that this might thus recursively instantiate
        references and thus modify the .references of (indirectly) referenced SqlModelBuilder
        instances.

        :raise Exception: If there are cyclic references between the recursively referenced
            SqlBuilderModels
        """
        # Note: Although it is not possible to create cycles in the SqlModel graph (see docstring
        #   SqlModel), it is possible to create cycles between SqlModelBuilder instances. Python will raise
        #   an error when it sees unbounded recursion, so we actually don't need to check for that here.
        for reference_name in self._references.keys():
            reference_value = self._references[reference_name]
            if not isinstance(reference_value, SqlModel):
                if isinstance(reference_value, SqlModelBuilder):
                    self._references[reference_name] = reference_value.instantiate_recursively()
                else:
                    raise Exception(f'In class {self.__class__.__name__} the reference {reference_name} '
                                    f'is not an SqlModel, but of '
                                    f'type {type(reference_value)}.')
        return self.instantiate()

    def __call__(self: TB, **values) -> 'SqlModel[TB]':
        self.set_values(**values)
        return self.instantiate()

    def instantiate(self: TB) -> 'SqlModel[TB]':
        """
        Create an instance of SqlModel[TB] based on the placeholders, references,
        materialization, and placeholders_to_sql of self.

        If the exact same instance (as determined by result.hash) has been created already by this class,
        then that instance is returned and the newly created instance is discarded.
        """
        self._check_is_complete()
        instance = SqlModel(model_spec=self,
                            placeholders=self.placeholders,
                            references=self.references,
                            materialization=self.materialization,
                            materialization_name=self.materialization_name)
        # If we already once created the exact same instance, then we'll return that one and discard the
        # newly created instance.
        if instance.hash not in self._cache_created_instances:
            self._cache_created_instances[instance.hash] = instance
        return self._cache_created_instances[instance.hash]

    def set_values(self: TB, **values) -> TB:
        """
        Set values that can either be references or placeholders.
        If a value is a placeholder value then it must be immutable
        If a value is a reference then it must either be an instance of SqlModelBuilder or of SqlModel
        :param values:
        :return: self
        """
        for key, value in values.items():
            if key in self.spec_references:
                if not isinstance(value, (SqlModel, SqlModelBuilder)):
                    raise ValueError(f'reference ({key}) is of incorrect type: {type(value)}')
                self._references[key] = value
            elif key in self.spec_placeholders:
                # There is no straightforward way to check whether a value is immutable. However virtual
                # all immutable objects are also hashable, so we check that instead.
                if not isinstance(value, collections.abc.Hashable):
                    raise ValueError(f'placeholder ({key}) is of incorrect type: {type(value)}')
                self._placeholders[key] = value
            else:
                raise ValueError(f'Provided parameter {key} is not a valid placeholder nor reference for '
                                 f'class {self.__class__.__name__}. '
                                 f'Valid references: {self.spec_references}. '
                                 f'Valid placeholders: {self.spec_placeholders}.')
        return self

    def set_materialization(self: TB, materialization: Materialization) -> TB:
        """
        Set the materialization
        :return: self
        """
        self.materialization = materialization
        return self

    def set_materialization_name(self: TB, materialization_name: Optional[str]) -> TB:
        """
        Set the materialization_name
        :return: self
        """
        self.materialization_name = materialization_name
        return self

    def _check_is_complete(self):
        """
        Raises an Exception if either references or placeholder values are missing
        """
        self.assert_adheres_to_spec(references=self.references, placeholders=self.placeholders)


class SqlModel(Generic[T]):
    """
    An Immutable Sql Model consists of a sql select query, placeholder values and references to other
    Sql models.

    # Graphs of models
    The references to other models make it is possible to use the output of the sql queries of other models
    as the input for a new model. References link models in an directed-acyclic-graph, with the current
    model as starting point. If a model is referenced by other models, then its part of the graph of those
    models. A model will not be aware that it is part of another model's graph.

    # Immutability
    Instances of this class are immutable. This has a few desirable consequences:
        * A model can safely be used in another model's graph, as that other model can rely on the fact
            that the model won't change.
        * Each model calculates an md5-hash at initialization based on its own attributes and recursively
            referenced attributes. This hash only has to be calculated once, since all those (recursive)
            attributes are immutable. The hash can be used to identify identical models and used for
            optimizations when generating sql.
        * All references have to be set at initialization, the referenced objects have to be already
            initialized models, and references are unidirectional, therefore it is not possible to create
            cycles in the graph.
    """
    def __init__(self,
                 model_spec: T,
                 placeholders: Mapping[str, Hashable],
                 references: Mapping[str, 'SqlModel'],
                 materialization: Materialization,
                 materialization_name: Optional[str] = None
                 ):
        """
        :param model_spec: ModelSpec class defining the sql, and the names of the placeholders and references
            that this class must have.
        :param placeholders: Dictionary mapping placeholder names to placeholder values. Important: the
            values must be immutable. If not then the instance as a whole is not immutable, which will
            invalidate assumptions that code might hold about these instances. See the class's docstring
            for information on why this is important.
        :param references: Dictionary mapping reference names to instances of SqlModels.
        :param materialization: How this model should be materialized, e.g. as CTE, as a table, etc.
        :param materialization_name: If the materialization is a database object (e.g. table, view,
            temp table), then this value can be set to override the default name of the object.
        """
        self._model_spec = model_spec
        self._generic_name = model_spec.generic_name
        self._sql = model_spec.sql
        self._references: Mapping[str, 'SqlModel'] = references
        self._placeholders: Mapping[str, Any] = placeholders
        self._materialization = materialization
        self._materialization_name = materialization_name
        self._placeholder_formatter = model_spec.placeholders_to_sql

        # Verify completeness of this object: references and placeholders
        self._model_spec.assert_adheres_to_spec(references=self.references,
                                                placeholders=self.placeholders)
        # Calculate unique hash for this model's sql, placeholder values, materialization and references
        self._hash = self._calculate_hash()

    def _calculate_hash(self) -> str:
        """
        Calculate md5 hash of the immutable data of this model that will be used for sql generation by the
        sql_generator. The hash is unique for the combination of the following attributes, and as such is
        unique for the sql that will be generated.
        Attributes considered in hash:
            1. generic_name
            2. sql
            3. placeholder values
            4. materialization
            5. references, and recursively their generic_name, sql, placeholder values, materialization,
                materialization_name, and recursive references
            6. materialization_name
        :return: 32 character string representation of md5 hash
        """
        data = {
            'generic_name': self.generic_name,
            'sql': self.sql,
            'properties': self.placeholders_formatted,
            'materialization': self.materialization.type_name,
            'references': {
                ref_name: model.hash for ref_name, model in self.references.items()
            }
        }
        if self.materialization_name is not None:
            data['materialization_name'] = self.materialization_name
        data_bytes = repr(data).encode('utf-8')
        return hashlib.md5(data_bytes).hexdigest()

    @property
    def model_spec(self):
        return deepcopy(self._model_spec)

    @property
    def generic_name(self) -> str:
        """ Name for the type of model."""
        return self._generic_name

    @property
    def materialization_name(self) -> Optional[str]:
        """ Optional name for this specific instance of the model_spec that this model implements. """
        return self._materialization_name

    @property
    def sql(self) -> str:
        return self._sql

    @property
    def references(self) -> MutableMapping[str, 'SqlModel']:
        # return shallow-copy of the dictionary.
        # keys are strings and thus immutable, values are included uncopied.
        return {key: value for key, value in self._references.items()}

    @property
    def placeholders(self) -> MutableMapping[str, Hashable]:
        # return deepcopy of the dictionary
        return cast(MutableMapping[str, Hashable], deepcopy(self._placeholders))

    @property
    def materialization(self) -> Materialization:
        return self._materialization

    @property
    def hash(self) -> str:
        """
        Unique 32 character hash based on this object's attributes and attributes of referenced models.
        If two instances have the same hash, they are equal for all intents and purposes,
        see _calculate_hash() for more information.

        This is different from the __hash__() function, which returns an integer, and is much more likely
        to collide.
        """
        return self._hash

    @property
    def placeholders_formatted(self) -> Dict[str, str]:
        return self._placeholder_formatter(self._placeholders)

    def copy_override(
        self: TSqlModel,
        *,
        model_spec: T = None,
        placeholders: Mapping[str, Hashable] = None,
        references: Mapping[str, 'SqlModel'] = None,
        materialization: Materialization = None,
        materialization_name: Union[Optional[str], NotSet] = not_set
    ) -> TSqlModel:
        """
        Create a copy of this instance, with the specified fields updated.
        Any fields that are not set will get the current value in the returned copy.

        Note that as None is a valid value for materialization_name, therefore we use the special token
        `not_set` to mean "keep current value".
        """
        materialization_name_value = \
            self.materialization_name if materialization_name is not_set else materialization_name
        return self.__class__(
            model_spec=self.model_spec if model_spec is None else model_spec,
            placeholders=self.placeholders if placeholders is None else placeholders,
            references=self.references if references is None else references,
            materialization=self.materialization if materialization is None else materialization,
            materialization_name=materialization_name_value
        )

    def copy_set(self: TSqlModel, new_placeholders: Mapping[str, Any]) -> TSqlModel:
        """
        Return a copy with the given placeholder values of this model updated.

        deprecated:: Use copy_override() if possible.
            https://github.com/objectiv/objectiv-analytics/issues/412
        """
        placeholders = self.placeholders
        for new_key, new_val in new_placeholders.items():
            if new_key not in placeholders:
                raise ValueError(f'Trying to update non-existing placeholder key: {new_key}. '
                                 f'placeholder keys: {sorted(placeholders.keys())}')
            placeholders[new_key] = new_val
        return self.copy_override(placeholders=placeholders)

    def copy_link(self: TSqlModel, new_references: Dict[str, 'SqlModel']) -> TSqlModel:
        """
        Create a copy with the given references of this model updated.

        Take care to not create cycles in the reference graph when using this function. Generally when
        working with a full graph of models its best to use the wrapper methods in graph_operations.py

        deprecated:: Use copy_override() if possible.
            https://github.com/objectiv/objectiv-analytics/issues/412
        """
        references = self.references
        for new_key, new_val in new_references.items():
            if new_key not in references:
                raise ValueError(f'Trying to update non-existing references key: {new_key}. '
                                 f'Reference keys: {sorted(references.keys())}')
            references[new_key] = new_val
        return self.copy_override(references=references)

    def copy_set_materialization(self: TSqlModel, materialization: Materialization) -> TSqlModel:
        """
        Create a copy with the given materialization of this model updated.

        deprecated:: Use copy_override() if possible.
            https://github.com/objectiv/objectiv-analytics/issues/412
        """
        if self.materialization == materialization:
            return self
        return self.copy_override(materialization=materialization)

    def copy_set_materialization_name(self: TSqlModel, materialization_name: Optional[str]) -> TSqlModel:
        """
        Create a copy with the given materialization_name of this model updated.

        deprecated:: Use copy_override() if possible.
            https://github.com/objectiv/objectiv-analytics/issues/412
        """
        if self.materialization_name == materialization_name:
            return self
        return self.copy_override(materialization_name=materialization_name)

    def set(self: TSqlModel,
            reference_path: RefPath,
            **placeholders) -> TSqlModel:
        """
        Create a (partial) copy of the graph that can be reached from self, with the placeholder values of
        the referenced node updated.

        The node identified by the reference_path is copied and updated, as are all nodes that
        (indirectly) refer that node. The updated version of self is returned.

        This instance, and all nodes that it refers recursively are unchanged.
        :param reference_path: references to traverse to get to the node that has to be updated
        :param placeholders: placeholder values to update
        :return: an updated copy of this node
        """
        # import locally to prevent cyclic imports
        from sql_models.graph_operations import get_node, replace_non_start_node_in_graph
        if reference_path == tuple():
            return self.copy_set(placeholders)
        replacement_model = get_node(self, reference_path).copy_set(placeholders)
        return replace_non_start_node_in_graph(self, reference_path, replacement_model)

    def link(self: TSqlModel,
             reference_path: RefPath,
             **references) -> TSqlModel:
        """
        Create a (partial) copy of the graph that can be reached from self, with the references of the
        referenced node updated.

        The node identified by the reference_path is copied and updated, as are all nodes that
        (indirectly) refer that node. The updated version of self is returned.

        This instance, and all nodes that it refers recursively are unchanged.
        :param reference_path: references to traverse to get to the node that has to be updated
        :param references: references to update
        :return: an updated copy of this node
        """
        # import locally to prevent cyclic imports
        from sql_models.graph_operations import get_node, replace_non_start_node_in_graph
        if reference_path == tuple():
            return self.copy_link(new_references=references)
        replacement_model = get_node(self, reference_path).copy_link(references)
        result = replace_non_start_node_in_graph(self, reference_path, replacement_model)
        return result

    def set_materialization(self: TSqlModel,
                            reference_path: RefPath,
                            materialization: Materialization) -> TSqlModel:
        """
        Create a (partial) copy of the graph that can be reached from self, with the materialization of the
        referenced node updated.

        The node identified by the reference_path is copied and updated, as are all nodes that
        (indirectly) refer that node. The updated version of self is returned.

        This instance, and all nodes that it refers recursively are unchanged.
        :param reference_path: references to traverse to get to the node that has to be updated
        :param materialization: materialization value
        :return: an updated copy of this node
        """
        # import locally to prevent cyclic imports
        from sql_models.graph_operations import get_node, replace_non_start_node_in_graph
        if reference_path == tuple():
            return self.copy_set_materialization(materialization)
        replacement_model = get_node(self, reference_path).copy_set_materialization(materialization)
        return replace_non_start_node_in_graph(self, reference_path, replacement_model)

    def set_materialization_name(self: TSqlModel,
                                 reference_path: RefPath,
                                 materialization_name: Optional[str]) -> TSqlModel:
        """
        Create a (partial) copy of the graph that can be reached from self, with the materialization_name of
        the referenced node updated.

        The node identified by the reference_path is copied and updated, as are all nodes that
        (indirectly) refer that node. The updated version of self is returned.

        This instance, and all nodes that it refers recursively are unchanged.
        :param reference_path: references to traverse to get to the node that has to be updated
        :param materialization_name: materialization_name value
        :return: an updated copy of this node
        """
        # import locally to prevent cyclic imports
        from sql_models.graph_operations import get_node, replace_non_start_node_in_graph
        if reference_path == tuple():
            return self.copy_set_materialization_name(materialization_name)
        replacement_model = get_node(self, reference_path).copy_set_materialization_name(
            materialization_name)
        return replace_non_start_node_in_graph(self, reference_path, replacement_model)

    def __eq__(self, other) -> bool:
        """
        Two SqlModels are equal if they have the same unique hash, and the same placeholder_formatter.
        This means the SqlModels effectively will lead to the same sql code when compiled. And
        additionally, derived models (e.g. through .set()) will be equal too if they are derived in the
        same way.

        This equality check does not take into account whether the classes are of the same (sub)class
        type, or whether the model_spec is the same type. As that ultimately won't affect the generated
        sql.
        """
        if not isinstance(other, SqlModel):
            return False
        # There is one edge-case (other than incredible unlikely md5 hash-collisions) where comparing the
        # hash to determine equality is not satisfactory. If a model has a non-standard
        # self._placeholder_formatter. Then the following scenario is possible:
        #   a.hash == b.hash
        #   c, d = a.set(tuple(), placeholder='new value'), b.set(tuple(), placeholder='new value')
        #   c.hash != d.hash
        # The same operation on two equal objects should render two equal objects again. By also including
        # the _placeholder_formatter in the comparison we can guarantee that.
        return self.hash == other.hash and self._placeholder_formatter == other._placeholder_formatter

    def __hash__(self) -> int:
        """ python hash. Must not be confused with the unique hash that is self.hash """
        return hash(self.hash)


class SourceTableModelBuilder(SqlModelBuilder):
    """
    Builder that instantiates a SqlModel with SOURCE materialization to refer to source data tables
    """

    def __init__(self, name: str):
        super().__init__()
        self._sql = ''
        self._generic_name = name
        self.set_materialization_name(name)
        self.set_materialization(Materialization.SOURCE)

    @property
    def sql(self):
        return ''


class CustomSqlModelBuilder(SqlModelBuilder):
    """
    Builder that instantiates a SqlModel that can run custom sql and refer custom tables.
    """

    def __init__(self, sql: str, name: str = None):
        """
        :param sql: sql of the model
        :param name: optional override of the generic name (default: 'CustomSqlModel')
        """
        self._sql = sql
        if name:
            self._generic_name = name
        else:
            self._generic_name = 'CustomSqlModel'
        super().__init__()

    @property
    def sql(self):
        return self._sql

    @property
    def generic_name(self):
        return self._generic_name


def escape_raw_sql(sql: str) -> str:
    """
    Take any raw sql that will be used in a model and escape all '{' and '}'. This prevents the sql_generator
    from interpreting '{' and '}' as either placeholders or references.
    """
    return escape_format_string(sql, 2)


def escape_placeholder_value(value: str) -> str:
    """
    Take any value that will be used as a placeholder value, and escape all '{' and '}'. This prevents
    the sql_generator from interpreting '{' and '}' as references.
    """
    return escape_format_string(value, 1)


def escape_format_string(value: str, times=1) -> str:
    """
    Escape value for python's format() function. i.e. `escape_format_string(value).format() == value`.

    SqlModels.sql is formatted twice by the sql_generator.py's to_sql() function. Once to replace the
    placeholder values, and once to replace the references.
    So any raw sql that is part of a SqlModel.sql and contains either '{' or '}' should be escaped twice.
    Any values that get inserted by the first call to format (replacing the placeholders), should be escaped
    once.
    """
    for _ in range(times):
        value = value.replace('{', '{{').replace('}', '}}')
    return value



"""
Copyright 2022 Objectiv B.V.
"""
from typing import TYPE_CHECKING, Optional, Union, List, cast

import numpy
import pandas

from bach import SeriesAbstractNumeric, SeriesNumericInterval
from sql_models.util import is_bigquery

if TYPE_CHECKING:
    from bach.dataframe import DataFrame


class PlotHandler(object):
    """
    Preprocesses data needed for plotting and uses Pandas ``DataFrame.plot``.
    """
    df: 'DataFrame'

    def __init__(self, df: 'DataFrame') -> None:
        self.df = df

    def hist(
        self,
        by: Optional[Union[str, List[str]]] = None,
        bins: int = 10,
        **kwargs,
    ):
        """
        Draw a histogram representation of DataFrame's numeric columns.

        :param by: series to group data by. Currently, not supported
        :param bins: number of equal-width histogram bins.
        :param kwargs: additional keyword arguments supported by Pandas ``DataFrame.plot``

        :returns: a histogram plot (matplotlib.AxesSubplot)
        """

        if by:
            raise NotImplementedError('by is currently not supported.')

        from bach.series.series_numeric import SeriesAbstractNumeric
        numeric_columns = [s.name for s in self.df.data.values() if isinstance(s, SeriesAbstractNumeric)]

        if not numeric_columns:
            raise ValueError(
                "hist method requires numerical columns, nothing to plot."
            )

        freq_df = self._calculate_hist_frequencies(bins, numeric_columns)

        # prepare results for Pandas hist compatibility
        freq_df['lower_edge'] = cast(SeriesNumericInterval, freq_df['range']).lower
        freq_pdf = freq_df.to_pandas()
        freq_pdf = freq_pdf.pivot_table(
            columns='column_label',
            values='frequency',
            index='lower_edge',
            dropna=False,
            fill_value=0,
        )
        freq_pdf = freq_pdf.reset_index(level=-1, drop=False)

        # get lower bounds per range and add the last upper bound (last_bin + bin_width)
        bin_edges = freq_pdf['lower_edge'].to_numpy()
        bin_width = numpy.ediff1d(bin_edges)[-1]
        bin_edges = numpy.append(bin_edges, bin_edges[-1] + bin_width)

        # use calculated frequencies as weights, since Pandas will try to recalculate frequencies
        hist_data = pandas.DataFrame(data={col: freq_pdf['lower_edge'] for col in numeric_columns})
        weights = freq_pdf[numeric_columns].to_numpy()
        return hist_data.plot.hist(bins=bin_edges, weights=weights, **kwargs)

    def _calculate_hist_frequencies(self, bins: int, numeric_columns: List[str]) -> 'DataFrame':
        """
        Helper for creating histogram's value frequencies per bin.

        returns a DataFrame containing the following data columns:
            * column_label (names of numeric columns)
            * range (range of each bin)
            * frequency (number of values the range contains for the column label)
        """
        from bach.operations.cut import CutOperation

        df = self.df[numeric_columns].reset_index(drop=True)
        # stack the df in order to have all numeric values in a single series
        label_values_series = cast(SeriesAbstractNumeric, df.stack())
        bins_per_col_df = CutOperation(
            label_values_series,
            bins=bins,
            method='bach',
            include_empty_bins=True,
            ignore_index=False,
        )().to_frame().reset_index(drop=False)

        # create frequency distribution dataframe per label (numeric column)
        # labels are contained in __stacked_index series (result from DataFrame.stack)
        frequencies = bins_per_col_df.reset_index().groupby(by=['__stacked_index', 'range']).count()
        frequencies = frequencies.reset_index(drop=False)

        # rename columns to meaningful names
        frequencies = frequencies.rename(
            columns={'__stacked_index': 'column_label', '__stacked_count': 'frequency'},
        )
        frequencies['column_label'] = frequencies['column_label'].fillna('empty_bins')
        return frequencies[['column_label', 'frequency', 'range']]



"""
Copyright 2022 Objectiv B.V.
"""
from enum import Enum
from typing import Optional, Union, Sequence, List, Dict
from bach import DataFrame, Series
from bach.series.series import WrappedPartition


class ValuePropagationMethod(Enum):
    BACK_FILL = 'bfill'
    FORWARD_FILL = 'ffill'

    @classmethod
    def get_method(cls, method: str) -> 'ValuePropagationMethod':
        if method in ('ffill', 'pad'):
            return cls.FORWARD_FILL

        if method in ('bfill', 'backfill'):
            return cls.BACK_FILL

        raise ValueError(f'"{method}" is not a valid propagation method.')


class ValuePropagation:
    """
    In order to instantiate this class you should provide the following params:
    df: a DataFrame containing all series to be filled by value propagation
    method: a string identifying the ValuePropagationMethod to be used.
        - 'ffill' or 'pad': Fill missing values by propagating the last non-nullable value in each series.
        - 'bfill' or 'backfill: Fill missing values by using the next non-nullable value in each series.
    returns a new DataFrame with filled missing values.

    Steps of how values are propagated:
    1.  Number each row of the sorted DataFrame. The main reason of numbering each row, is because
        bfill is the opposite to ffill, so we can say that it uses a reversed sorting. The issue with this
        is that Bach will always sort NULLS LAST, therefore reversing the order by won't work. By numbering
        each row, we can at least rely on a non-nullable value.

    2. Sort dataframe based on the row number.
            Ascending if method == 'ffill' and descending if method == 'bfill'.
    3. Partition each series to be filled, a partition is the cumulative sum of the total amount
        of observed non-nullable values in the series.
    4. Group the series to be filled by its partition and assign the first value.
    """
    ROW_NUMBER_SERIES_NAME = '__row_number'

    def __init__(self, df: DataFrame, method: str):
        self._df = df
        self._method = ValuePropagationMethod.get_method(method)

    def propagate(
            self,
            sort_by: Optional[Union[str, Sequence[str]]] = None,
            ascending: Union[bool, List[bool]] = True,
            window: Optional[WrappedPartition] = None
    ):
        df = self._df.copy()

        if df.group_by:
            raise ValueError('ffill/bfill cannot be applied on grouped DataFrame. '
                             'please materialize first.')

        if sort_by:
            df = df.sort_values(by=sort_by, ascending=ascending)

        if not df.order_by:
            raise Exception('dataframe must be sorted in order to apply forward or backward fill.')

        # Number each sorted row this way we avoid issues when sorting NULL values
        # since bfill is ffill with reversed sort (this might generate different results)
        base_series = df[df.data_columns[0]]
        # order nulls last in window function, else numbering will be wrong
        window = window or df.groupby().window(na_position='last')
        window_group = list(window.index.keys())

        df[self.ROW_NUMBER_SERIES_NAME] = base_series.window_row_number(window)
        df = df.materialize(node_name='numbered_fillna')

        # sort values by row number
        by = window_group + [self.ROW_NUMBER_SERIES_NAME]
        ascending = self._method == ValuePropagationMethod.FORWARD_FILL

        df = df.sort_values(by=by, ascending=ascending)
        df = self._add_partition_per_data_columns(df, window_group=window_group)

        # sorting is lost due to materialization, still requires it
        df = df.sort_values(by=by, ascending=ascending)
        df = self._propagate_first_value_per_partition(df, window_group=window_group)

        # sort values, this way we respect the provided order by
        df = df.sort_values(by=self.ROW_NUMBER_SERIES_NAME)

        df = df[self._df.data_columns].materialize(node_name='nafilled')

        gb: List[Union[str, Series]] = []
        if window_group:
            if isinstance(window_group, str):
                gb = [window_group]
            elif isinstance(window_group, list):
                gb = window_group

        return df.drop(columns=gb)

    def _add_partition_per_data_columns(
        self,
        df: DataFrame,
        window_group: Optional[Union[str, Sequence[str]]] = None
    ) -> DataFrame:
        """
        Helper function that creates a partition column for each series to be filled
        this column contains the cumulative sum of the total amount of observed non-nullable values
        till the current row.
                Example:
                |   A  | __partition_A |
                |:----:|:-------------:|
                |   1  |       1       |
                | NULL |       1       |
                | NULL |       1       |
                |   3  |       2       |
                | NULL |       2       |
                |   4  |       3       |

        Returns a DataFrame containing a partition series per data column
        """
        from bach.partitioning import WindowFrameMode
        df_cp = df.copy()
        partition_window = df_cp.groupby(window_group).window(mode=WindowFrameMode.ROWS)
        for series_name, series in self._df.data.items():
            partition_name = f'__partition_{series.name}'
            df_cp[partition_name] = 1
            df_cp.loc[series.isnull(), partition_name] = 0
            df_cp[partition_name] = df_cp[partition_name].sum(partition=partition_window)

        return df_cp.materialize(node_name='fillna_partitioning')

    def _propagate_first_value_per_partition(self,
                                             df: DataFrame,
                                             window_group: Optional[Union[str, Sequence[str]]] = None):
        """
        Helper function that fills null values based on partitions
        NULL records are grouped by the partition and filled with the non-nullable value
        respective to the partition
              Example:
              |   A  | __partition_A | filled_A |
              |:----:|:-------------:|:--------:|
              |   1  |       1       |     1    |
              | NULL |       1       |     1    |
              | NULL |       1       |     1    |
              |   3  |       2       |     3    |
              | NULL |       2       |     3    |
              |   4  |       3       |     4    |

        Returns a DataFrame where all data columns are filled
        """
        # fill gaps with the first_value per partition
        df_cp = df.copy()

        gb: List[Union[str, Series]] = []
        if window_group:
            if isinstance(window_group, str):
                gb = [window_group]
            elif isinstance(window_group, list):
                gb = window_group

        initial_columns = self._df.data_columns
        fill_columns = [x for x in initial_columns if x not in gb]

        for series_name in fill_columns:
            partition_series = [df_cp[f'__partition_{series_name}']]
            gb_final = gb + partition_series
            window_first_value = df_cp.groupby(gb_final).window()
            df_cp[series_name] = df_cp[series_name].window_first_value(window_first_value)

        return df_cp.materialize(node_name='fillna_propagated_values')
